=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed April 2015 by Fredo6

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for any purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_Lib6DeepSelector.rb
# Original Date	:   27 Mar 15
# Description	:   Selector of grouponents at deep level
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module Traductor

T6[:MSG_DeepSelectorInvalidRed] = "Objects in Red will be unselected"

#=============================================================================================
#=============================================================================================
# Class DeepSelector: Manage interactive selection at any hierarchical level
#=============================================================================================
#=============================================================================================

class DeepSelector

#INIT: Instance initialization
def initialize__(*hargs)	
	puts "*****DEEEP SELECTOR INIT"
	#Parsing the arguments
	hargs.each { |arg| arg.each { |key, value|  parse_args(key, value) } if arg.class == Hash }	
	
	#Basic initialization
	@model = Sketchup.active_model
	@view = @model.active_view
	@tr_id = Geom::Transformation.new
	@ph = @view.pick_helper
	@ogl = Traductor::OpenGL_6.new

	#Instantiating the PseudoComp manager
	@pcompman = Traductor::PseudoCompManager.instantiate unless @pcompman

	#Reset environment
	init_colors
	init_texts
	reset
end

#INIT: Parse the arguments of the initialize method
def parse_args(key, value)
	skey = key.to_s
	case skey
	when /keyman/i
		@keyman = value
	when /pcompman/i
		@pcompman = value
	when /notify_proc/i
		@notify_proc = value
	end	
end	

#INIT: Notify the caller about events
def notify_caller(event, *args)	
	@notify_proc.call(event, *args) if @notify_proc
end

#INIT: Initialize the colors
def init_colors
	fac = 1.6
	@colors_wf = ['yellow', 'gold', 'goldenrod', 'orange', 'darkorange', 'sandybrown']
	@colors_wf_inst = @colors_wf.collect { |color| G6.color_light_dark(color, fac) }
	@ncolors_wf = @colors_wf.length
	@color_selected = 'blue'
	@color_selected_inst = Sketchup::Color.new 100, 100, 255
	@color_induced = 'green'
	@color_highlight = 'deepskyblue'
	@color_highlight_inst = 'skyblue'
	@color_invalid = 'red'
	@color_drop_unselect = 'pink'
end

#INIT: Initialize the text
def init_texts
	@wid_dropdown = 200
	@msg_unselect = T6[:T_BUTTON_UnSelect]
	@msg_group = T6[:T_TXT_Group]
	@msg_comp = T6[:T_TXT_Component]
	@msg_invalid = T6[:MSG_DeepSelectorInvalidRed]
end
	
#INIT: Reset the sequence editor parameters
def reset
	puts "DEEP Reset"
	@hsh_selected_comps = {}
	@hsh_induced_comps = {}
	@last_comp_picked = nil
	reset_drawing
end

def reset_drawing
	@hsh_wireframes = {}
	@hsh_comp_invalid = {}
end

#INIT: Activate the tool
def activate
	reset
	@invoked = true
end

def deactivate
	@invoked = false
end

#-----------------------------------------------
# TOP: Top level methods
#-----------------------------------------------

#KEY: Manage key events
def key_manage(event, key, view)
	#puts "Deep keyman #{event} key = #{key}"
	
	case event
	
	#SHIFT key
	when :shift_down
		
	when :shift_toggle
		if @mode == :origin
			
		else	
			
		end	
	when :shift_up
		
	
	#CTRL key	
	when :ctrl_down	
		#handle_move_zero
		
	when :ctrl_up

	when :ctrl_other_up
	
	#ALT key
	when :alt_down
		
	when :alt_up	
		
		
	#Arrows	
	when :arrow_down	
		
	
	#Backspace	
	when :backspace
		

	#TAB	
	when :tab
		
	when :shift_tab
		
	#Otherkeys
	else
		return false
	end
	true
end

#TOP: Get the number of objects selected
def get_selected_objects ; @hsh_selected_comps ; end
def get_selected_objects_id ; @hsh_selected_comps.keys ; end

#-----------------------------------------------
# SELECTION: Manage selection
#-----------------------------------------------
	
def transfer_preselection
	save_previous_selection
end

def cancel_selection
	if @palette.dropdown_active?
		@palette.dropdown_finish
		return true
	end
	save_previous_selection
	unselect_all
	false
end

def save_previous_selection
	puts "SAVE SELECTION SELECTOR empty = #{@hsh_selected_comps.empty?}"
	@hsh_selected_comps_previous = @hsh_selected_comps unless @hsh_selected_comps.empty?
end

def restore_previous_selection(no_notify=false)
	puts "rstpre previous previous = #{@hsh_selected_comps_previous && @hsh_selected_comps_previous.length > 0}"
	if @hsh_selected_comps_previous && @hsh_selected_comps_previous.length > 0
		@hsh_selected_comps = @hsh_selected_comps_previous
	end	
	update_induced
	notify_caller :selection_restore
end

#SELECT: Un-register all selection
def unselect_all
	@hsh_selected_comps = {}	
	@hsh_induced_comps = {}	
	@last_comp_picked = nil
	notify_caller :selection_cleared
end

#SELECT: Check if current selection is empty
def selection_empty?
	@hsh_selected_comps.empty?
end

#SELECT: Check if there is a previous selection
def selection_previous?
	@hsh_selected_comps_previous && !@hsh_selected_comps_previous.empty?
end

#SELECT: Check if an object is already part of the selection
def is_part_of_selection?(comp)
	return false unless comp
	@hsh_selected_comps.has_key?(comp.entityID)
end

#SELECT: Upload a selection
def upload_selection(lst_comp_info)
	#Reset the environment
	reset
	
	#Load the selection and calculating the top transformation
	lst_comp_info.each do |comp_info|
		comp, chain_comp = comp_info
		puts "UPLOAD SEL #{comp_info.inspect}"
		tr = @tr_id
		chain_comp.each do |c|
			tr = c.transformation * tr
		end
		@hsh_selected_comps[comp.entityID] = [comp, tr, chain_comp]
	end
	
	#Saving the selection
	save_previous_selection
	
	#Notifying the caller
	notify_caller :selection_uploaded
end

#-----------------------------------------------
# MOVE: Mouse move methods
#-----------------------------------------------

#MOVE: Detect elements and chain of grouponents under the mouse
def chain_under_mouse(view, x, y)	
	#Initialization
	@in_the_void = true

	#Picking the point
	@ph.do_pick x, y
	
	#Getting the picked element
	return nil unless @ph.picked_face || @ph.picked_edge || @ph.picked_element
	@in_the_void = false
	
	#Calculating the parent chain
	chain = @ph.path_at(0).find_all { |e| e.instance_of?(Sketchup::Group) || e.instance_of?(Sketchup::ComponentInstance) }
	
	#Combining info with transformation
	chain_info = []
	t = @tr_id
	chain.each do |comp|
		next if exclude_comp?(comp)
		t = t * comp.transformation
		#chain_info.push [comp, t, comp.entityID]
		chain_info.push [comp, t]
	end
	
	#Returning the chain	
	chain_info.reverse
end
	
#MOVE: Notify the caller for filtering comps	
def exclude_comp?(comp)
	notify_caller :exclude_comp, comp
end
	
def filter_comps
	@hsh_selected_comps.each do |comp_id, a|
		comp, = a
		@hsh_selected_comps.delete comp_id if exclude_comp?(comp)
	end	
	notify_caller :selection_changed
end
	
#MOVE: top level move method
def handle_move_zero ; handle_move(0, @xmove, @ymove, @view) if @xmove ; end
def handle_move(flags, x, y, view)
	@xmove = x
	@ymove = y
	
	#Selection
	analyze_chain_info(view, x, y) unless @palette.dropdown_active?
	
	#Hover mode
	if @keyman.shift_down?
		puts "shift down"
	end
end

#-----------------------------------------------
# CLICK: Mouse click methods
#-----------------------------------------------

#CLICK: Handle Mouse DOWN
def handle_button_down(flags, x, y, view)	
	puts "DEEP SELECT DOWN SELECTOR"
	@time_down_start = Time.now
	@xdown = x
	@ydown = y
	@action_stopped = false
	
	#Selection mode - Actioning the drop down list
	if @chain_info
		compute_dropdown
		@palette.dropdown_invoke @flpal_seq_obj
	end
end

#CLICK: Handle Mouse UP
def handle_button_up(flags, x, y, view)
	#if !@chain_info
	if !chain_under_mouse(view, x, y)
		puts "UP in void"
		#if @hsh_selected_comps.empty?
			notify_caller :click_in_void
		#end
		unselect_all unless @palette.dropdown_active?
		@palette.dropdown_finish
	elsif Time.now - @time_down_start < 0.4 && (x - @xdown).abs < 3 && (y - @ydown).abs < 3 && @lst_obj_levels && @lst_obj_levels.length <= 1
		select_objects
		@palette.dropdown_finish
	end	
end

#CLICK: Handle double click
def handle_doubleclick(flags, x, y, view)
	@palette.dropdown_finish
	return false unless @chain_info
	select_objects(0, true)
	true
end

#-----------------------------------------------
# SELECT: Handle selection of Objects
#-----------------------------------------------

#SELECT: Get the chain of selected comps
def chain_comp_info
	return [] unless @chain_info && @chain_info.length > 0
	@chain_info.collect { |a| [a[0], a[1]] }
end

#SELECT: Analyze the chain info under mouse
def analyze_chain_info(view, x, y)
	@hsh_comp_invalid = {}
	@icomp_hover = nil
	@chain_info = chain_under_mouse(view, x, y)	
	if @chain_info
		isel0 = 0
		@isel_comp = check_super_levels(isel0)
		#comp0, tr0, key0 = @chain_info[@isel_comp]
		comp0, tr0 = @chain_info[@isel_comp]
		if comp0
			if comp0 != @last_comp_picked
				@last_comp_picked = nil
				#@hsh_comp_invalid[key0] = true if is_part_of_selection?(comp0)
				@hsh_comp_invalid[comp0.entityID] = true if is_part_of_selection?(comp0)
			end
			lembed = check_embedded(comp0)
			lembed.each { |comp| @hsh_comp_invalid[comp.entityID] = true }
		end	
	else
		@last_comp_picked = nil
	end	
end

#SELECT: Calculate the drop down list
def compute_dropdown
	return false unless @chain_info
	
	#Building the drop down list 
	@lst_obj_levels = []
	isel = @isel_comp
	for i in isel..@chain_info.length-1	
		comp, tr = @chain_info[i]
		j = i - isel
		if j == 0 && is_part_of_selection?(comp)
			label = @msg_unselect
			color = @color_drop_unselect
		else	
			label = j.to_s
			color = (j == 0) ? @color_highlight : @colors_wf[(j-1).modulo(@ncolors_wf)]
		end	
		name = comp.name
		if name.empty?
			name = (comp.instance_of?(Sketchup::Group)) ? @msg_group : @msg_comp
		end	
		@lst_obj_levels.push [label, name, G6.abbreviate_name(name, @wid_dropdown), color]
	end	
	true
end

#SELECT: Select an object
def select_objects_from_dropdown(ilevel)
	select_objects @isel_comp + ilevel
	@palette.dropdown_finish
end

#SELECT: Select an object
def hover_objects_from_dropdown(ilevel)
	@icomp_hover = nil
	return unless @chain_info
	info = @chain_info[@isel_comp + ilevel]
	return unless info
	@icomp_hover = info[0]
	if @icomp_hover
		@hsh_comp_invalid = {}
		lembed = check_embedded(@icomp_hover)
		lembed.each { |comp| @hsh_comp_invalid[comp.entityID] = true }
		@hsh_induced_comps = @pcompman.joint_compute_impacted(@icomp_hover)
	end
end

#SELECT: Select an object
def select_objects(isel=0, forced=false)
	puts "Deep select objects"
	unless @chain_info
		unselect_all
		return 
	end	
	
	#Checking if comps higher in the chain are already selected. If so, unselect them
	isel = check_super_levels(isel)
	
	comp0, tr0 = @chain_info[isel]
	if !forced && is_part_of_selection?(comp0)
		unregister_comp(comp0)
		notify_caller :selection_changed
		return
	end	
	
	#Checking if selected comp contains selected comps
	lembed = check_embedded(comp0)
	lembed.each { |comp| unregister_comp(comp) }

	#Registering and selecting the pseudocomp
	register_comp(@chain_info, isel)
	
	#Transfer the selection
	transfer_preselection
	
	#Change of selection
	notify_caller :selection_changed
end

#SELECT: Register an object in the sequence
def register_comp(chain_info, isel)
	comp, tr = @chain_info[isel]
	return unless comp
	chain_comp = chain_info[isel..-1].collect { |a| a[0] }
	@hsh_selected_comps[comp.entityID] = [comp, tr, chain_comp]
	update_induced
	@last_comp_picked = comp
	
	puts "SELECT CHAIN ID = #{chain_comp.inspect}"
end

def update_induced
	@hsh_induced_comps = @pcompman.joint_compute_impacted(@hsh_selected_comps.keys)
end

#SELECT: Un-register an object in the sequence
def unregister_comp(comp)
	return unless comp
	@hsh_selected_comps.delete comp.entityID
	update_induced
	@last_comp_picked = nil
end

#SELECT: Return the higher level of current selection
def check_super_levels(isel)
	lerase = []
	for i in isel..@chain_info.length-1
		comp, = @chain_info[i]
		return i if is_part_of_selection?(comp)		
	end
	isel
end

#SELECT: List sub-objects in the current picked object
def check_embedded(comp0, lerase=nil)
	return [] if @hsh_selected_comps.empty?
	lerase = [] unless lerase
	entities0 = G6.grouponent_entities(comp0)
	lcomp0 = entities0.grep(Sketchup::Group) + entities0.grep(Sketchup::ComponentInstance)
	@hsh_selected_comps.each do |id, a|
		comp, = a
		lerase.push comp if lcomp0.include?(comp)
	end
	lcomp0.each do |comp0|
		lerase = check_embedded(comp0, lerase)
	end	
	lerase
end	

#-----------------------------------------------
# DRAW: Handle drawing for the sequence editor
#-----------------------------------------------

#DRAW: top level drawing method
def draw(view)	
	#Drawing the current chain
	@factor_box = 1.05
	drawing_chain_info(view, @chain_info, @isel_comp) if @chain_info
end

#DRAW: Drawing message
def post_draw(view)			
	drawing_message_invalid view
end

#DRAW: Draw a chain of grouponent
def drawing_chain_info(view, chain_info, isel)	
	return unless chain_info
	
	#Drawing the selected comp
	comp, tr, key = chain_info[isel]
	comp_tr = [comp, tr]
	if @hsh_comp_invalid[key]
		color = @color_invalid
		color_inst = nil
	elsif comp == @last_comp_picked 	
		color = @color_selected
		color_inst = @color_selected_inst
	else
		color = @color_highlight
		color_inst = @color_highlight_inst
	end
	drawing_comp_wireframe(view, comp_tr, color, color_inst)
	G6.grouponent_draw_box view, comp_tr, color, color_inst, 3, nil, @factor_box
	
	#Drawing the box for chain of parents
	for i in isel+1..chain_info.length-1	
		comp, tr = chain_info[i]
		comp_tr = [comp, tr]
		color = @colors_wf[(i-isel-1).modulo(@ncolors_wf)]
		color_inst = @colors_wf_inst[(i-isel-1).modulo(@ncolors_wf)]
		wid = (comp == @icomp_hover) ? 3 : 1
		G6.grouponent_draw_box view, comp_tr, color, color_inst, wid, nil, @factor_box
	end	
	
	#Drawing the induced comps
	drawing_induced(view)
end

#DRAW: Draw the current selection of objects
def drawing_selection(view, factor_box=nil)
	return if @hsh_selected_comps.empty?
	factor_box = 1.05 unless factor_box
	@hsh_selected_comps.each do |id, comp_tr|
		comp, tr = comp_tr
		if @hsh_comp_invalid[id]
			color = @color_invalid
			color_inst = nil
		else
			color = @color_selected
			color_inst = @color_selected_inst
		end
		#lst_tr = G6.grouponent_tr_top_list(comp)
		drawing_comp_wireframe(view, comp_tr, color, color_inst)
		G6.grouponent_draw_box view, comp_tr, color, color_inst, 3, nil, factor_box
	end
	
	#Drawing the induced comps
	drawing_induced(view)	
end

#DRAW: Draw induced comps
def drawing_induced(view)
	return unless @hsh_induced_comps && @hsh_induced_comps.length > 0
	@hsh_induced_comps.each do |id, pcomp|
		next if @hsh_selected_comps[id]
		comp = pcomp.comp
		color = @color_induced
		wid = 2
		G6.grouponent_draw_box view, comp, color, nil, wid, '', @factor_box		
	end
end

#DRAW: Draw the wireframe for a grouponent
def drawing_comp_wireframe(view, comp_tr, color_main, color_inst)
	comp, tr = comp_tr
	return unless comp
	comp_id = comp.entityID
	lines_main, lines_inst = wireframe_compute(comp_tr)
	return unless lines_main
	view.line_stipple = ''
	view.line_width = 2
	color_inst = color_main unless color_inst
	view.drawing_color = color_main
	view.draw GL_LINES, lines_main.collect { |pt| G6.small_offset(view, pt) } if lines_main && lines_main.length > 0
	view.drawing_color = color_inst
	view.draw GL_LINES, lines_inst.collect { |pt| G6.small_offset(view, pt) } if lines_inst && lines_inst.length > 0
end

#DRAW: Compute the wireframe for selected components
def wireframe_compute(comp_tr)
	comp, tr_top = comp_tr
	return nil unless comp
	comp_id = comp.entityID
	lines = @hsh_wireframes[comp_id]
	return lines if lines
	
	lines_main = []
	lines_other = []
	@hsh_wireframes[comp_id] = [lines_main, lines_other]
	
	G6.grouponent_tr_top_list(comp).each do |tr|
		#if !tr_top || G6.transformation_equal?(tr_top * comp.transformation, tr)
		if !tr_top || G6.transformation_equal?(tr_top, tr)
			wireframe_compute_sub(comp, tr, lines_main)
		else
			wireframe_compute_sub(comp, tr, lines_other)
		end			
	end	

	@hsh_wireframes[comp_id]
end

#DRAW: Compute the wireframe for child of selected components
def wireframe_compute_sub(comp, tr, lines)
	entities = G6.grouponent_entities(comp)
	edges = entities.grep(Sketchup::Edge)
	edges.each { |e| lines << tr * e.start.position ; lines << tr * e.end.position }
	(entities.grep(Sketchup::Group) + entities.grep(Sketchup::ComponentInstance)).each do |g|
		wireframe_compute_sub(g, tr * g.transformation, lines)
	end
	lines
end

#DRAW: Drawing the invalid message for objects that will be unselected
def drawing_message_invalid(view)
	return if @hsh_comp_invalid.empty?
	G6.draw_rectangle_text(view, @xmove, @ymove, @msg_invalid, 'pink', 'red')
end

#-----------------------------------------------
# PALETTE: Palette Contribution for dropdown
#-----------------------------------------------
	
#PALETTE: Declare the drop down palette
def palette_contribution(prefix)
	@palette = Traductor::Palette.get_palette
	puts "PALETTE DEEP = #{@palette}"
	
	#Parameters
	@prefix = prefix
	@nb_sepa = 0
	@wid_sepa = 8
	
	#Drop down for object selection
	palette_dropdown_contribution
end

def palette_dropdown_contribution
	@nb_seq_obj = 15

	#Declare the floating palette
	@flpal_seq_obj = "#{@prefix}_Deep_Selector_pal_floating".intern
	hsh = { :dropdown_proc => true, :bk_color => 'lightgreen', :fr_color => 'navy' }
	@palette.declare_floating @flpal_seq_obj, hsh

	widtot = 200
	h = 20
	hshc = { :floating => @flpal_seq_obj, :width => widtot, :height => h, :justif => 'LH' }
	
	#List for levels of objects
	for i in 0..@nb_seq_obj-1
		palette_dropdown_level_button(i, hshc)
	end	
end

#PALETTE: declare buttons for the list of movements
def palette_dropdown_level_button(i, hshc)
	hidden_proc = proc { !@lst_obj_levels || !@lst_obj_levels[i] }
	text_proc = proc { "#{@lst_obj_levels[i][0]} - #{@lst_obj_levels[i][2]}" }
	tip_proc = proc { (@lst_obj_levels[i]) ? "#{@lst_obj_levels[i][1]}" : "" }
	bk_color_proc = proc { @lst_obj_levels[i][3] }
	hover_proc = proc { hover_objects_from_dropdown i }
	
	hsh = { :hidden_proc => hidden_proc, :text_proc => text_proc, :row => 0, :rank => @nb_seq_obj-i-1,
	        :tip_proc => tip_proc, :bk_color => bk_color_proc, :hover_proc => hover_proc }
	@palette.declare_button("#{@flpal_seq_obj}_#{i}".intern, hshc, hsh) { select_objects_from_dropdown(i) }	
end

#--------------------------------------------------
# MENU: Contextual menu
#--------------------------------------------------

#Contextual menu for list of grouponents and actions
def menu_contribution(cxmenu)
	return unless compute_dropdown
	cxmenu.add_sepa
	
	@lst_obj_levels.each_with_index do |a, i|
		label, name = a
		text = "#{label} - #{name}"
		cxmenu.add_item(text) { select_objects_from_dropdown(i) }
	end	
end

end	#class DeepSelector

end	#End Module Traductor