=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed by Fredo6 - Copyright July 2015

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_Lib6Marshall.rb
# Original Date	:   14 Jul 2015
# Description	:   Marshalling of data as text (implementation)
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end


module Traductor

#--------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------
# class Marshall: Management of text marshalling
#--------------------------------------------------------------------------------------------------------------			 
#--------------------------------------------------------------------------------------------------------------			 

class Marshall

@@count = 0 unless defined?(@@count)

#Initialize the Marshall instance
def initialize__
	@pcompman = Traductor::PseudoCompManager.instantiate
	@pat = /(\{([A-Z][A-Z])([0-9]+)\})(.*)\1/
	@sub_return = "|||N|||"
end

#Get a new tag for beg and end
def tag(prefix, val)
	@@count += 1 
	tag = "{#{prefix}#{@@count}}"
	tag + val.to_s + tag
end

#Construct a separator for list and hash
def sepa(num=nil)
	(num) ? ",,,#{num},,," : ",,,#{@@count+1},,,"
end
	
#ENCODE the value <a> as a string	
def encode(a)
	aclass = a.class
	
	#Float, Integer and Length
	if aclass == Float
		s = tag(:FF, a)
	elsif aclass == Fixnum
		s = tag(:IN, a)
	elsif aclass == Length
		s = tag(:LN, a)
		
	#String and Symbol	
	elsif aclass == String
		s = tag(:SS, a.gsub("\n", @sub_return))
	elsif aclass == Symbol
		s = tag(:SY, a)
		
	#Hash Array	
	elsif aclass == Hash
		ls = []
		a.each { |key, val| ls.push encode(key) + "=>" + encode(val) }
		s = tag(:HH, ls.join(sepa))
		
	#Array	
	elsif aclass == Array
		ls = a.collect { |val| encode(val) }
		s = tag(:LL, ls.join(sepa))
		
	#Point3d, Vector3d, Transformation	
	elsif aclass == Geom::Point3d
		ls = a.to_a.collect { |x| encode(x) }
		s = tag(:PT, ls.join(sepa) )
	elsif aclass == Geom::Vector3d
		ls = a.to_a.collect { |x| encode(x) }
		s = tag(:VC, ls.join(sepa) )
	elsif aclass == Geom::Transformation
		ls = a.to_a.collect { |x| encode(x) }
		s = tag(:TT, ls.join(sepa) )
		
	#Sketchup Entities	
	elsif aclass == Sketchup::Group || aclass == Sketchup::ComponentInstance || 
	      aclass == Sketchup::Page || a.class == Sketchup::Layer
		if a.valid?
			uid = @pcompman.get_uid(a.entityID)
			uid = a.entityID unless uid
			s = tag(:SU, uid)
		else
			s = tag(:NL, nil)
		end	
		
	#SU Camera	
	elsif aclass == Sketchup::Camera
		camera = a
		params = [camera.eye, camera.target, camera.yaxis, camera.perspective?, camera.fov]
		ls = params.collect { |x| encode(x) }
		s = tag(:CA, ls.join(sepa))
		
	#Nil and Boolean	
	elsif a == true
		s = tag(:BT, a)
	elsif a == false
		s = tag(:BF, a)
	elsif a == nil
		s = tag(:NL, a)
		
	#Other cases	
	else
		s = ""
	end
	s
end

#DECODE an encoded string
def decode(s)
	return nil unless s
	return s.collect { |u| decode(u) } if s.class == Array
	a = s.scan @pat
	tag, code, num, sval = a[0]
	
	case code
	
	#Nil and Boolean
	when "NL"
		res = nil
	when "BT"
		res = true
	when "BF"
		res = false
		
	#Float, Integer, Length	
	when "FF"
		res = sval.to_f
	when "IN"
		res = sval.to_i
	when "LN"
		res = sval.to_l
		
	#String and Symbol	
	when "SS"
		res = sval.gsub(@sub_return, "\n")
	when "SY"
		res = sval.intern
		
	#Hash array	
	when "HH"
		lsh = s.split sepa(num)
		res = {}
		lsh.each do |a|
			a.match(/\=>/)
			key, val = $`, $'
			res[decode(key)] = decode(val)
		end
		
	#Array	
	when "LL"
		res = sval.split(sepa(num)).collect { |a| decode(a) }
		
	#Point3d, Vector3 and Transformation	
	when "PT"
		ls = sval.split(sepa(num)).collect { |a| decode(a) }
		res = Geom::Point3d.new *ls
	when "VC"
		ls = sval.split(sepa(num)).collect { |a| decode(a) }
		res = Geom::Vector3d.new *ls
	when "TT"
		ls = sval.split(sepa(num)).collect { |a| decode(a) }
		res = Geom::Transformation.new ls
		
	#SU Camera	
	when "CA"
		ls = sval.split(sepa(num)).collect { |a| decode(a) }
		res = Sketchup::Camera.new *ls
		
	#Encoded Sketchup entities
	when "SU"
		res = @pcompman.decode_uid(sval)
		
	#Other cases	
	else
		res = nil
	end
	res
end
	
end	#class Marshall

end	#module Traductor

