=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed by Fredo6 - Copyright April 2009

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_Lib6Palette.rb
# Original Date	:   8 May 2009 - version 1.0
# Description	:   Interactive button palette management (implementation)
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end


module Traductor

T6[:Palette_TIP_TipZone] = "Information area"
T6[:Palette_TIP_MsgZone] = "Message area"
T6[:Palette_TIP_Pos] = "Position Top or Bottom"
T6[:Palette_TIP_Shrink] = "Shrink / Expand Palette"
T6[:Palette_TIP_MsgVisible] = "Show / Hide message area"
T6[:Palette_TIP_SideLeft] = "Position Left or Right"

T6[:T_TIP_Increment] = "Increment value"
T6[:T_TIP_Decrement] = "Decrement value"

T6[:T_TIP_EditionUndo] = "Undo change"
T6[:T_TIP_EditionRedo] = "Redo last action"
T6[:T_TIP_EditionReset] = "Cancel all changes"
T6[:T_TIP_EditionRestore] = "Restore all changes"

#--------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------
# Class Palette: Button Palette management
#--------------------------------------------------------------------------------------------------------------			 
#--------------------------------------------------------------------------------------------------------------			 

class Palette

#Structure for an individual button

attr_reader :top_pos, :shrinked, :side_left, :message_visi

@@persistence = nil
@@palette = nil

#----------------------------------------------------------------------------
# Managing default parameters
#----------------------------------------------------------------------------

#Saving parameters across sessions
def persistence_save
	@@persistence = {} unless @@persistence

	#Common to all tools
	hsh = @@persistence
	hsh["PalettePos"] = @top_pos
	hsh["PaletteShrinked"] = @shrinked	
	hsh["PaletteSideLeft"] = @side_left	
	hsh["PaletteMessageVisi"] = @message_visi	
end

#Restoring parameters across sessions
def persistence_restore
	hsh = @@persistence
	unless hsh
		hsh = @@persistence = {}
		hsh["PalettePos"] = MYDEFPARAM[:DEFAULT_PalettePos]
		hsh["PaletteShrinked"] = MYDEFPARAM[:DEFAULT_PaletteShrinked]
		hsh["PaletteSideLeft"] = MYDEFPARAM[:DEFAULT_PaletteSideLeft]
		hsh["PaletteMessageVisi"] = MYDEFPARAM[:DEFAULT_PaletteMessageZone]
	end
	@top_pos = hsh["PalettePos"]
	@shrinked = hsh["PaletteShrinked"]	
	@side_left = hsh["PaletteSideLeft"]	
	@message_visi = hsh["PaletteMessageVisi"]
	@retina_factor = Traductor.retina_factor
	retina_main_transfo	
end

#----------------------------------------------------------------------------
# Initialization and declarations
#----------------------------------------------------------------------------

def self.get_palette ; @@palette ; end

#Create a palette instance
def initialize__(*args)
	@@palette = self
	
	#SU context
	@model = Sketchup.active_model
	@view = @model.active_view
	@render = @model.rendering_options
	@tr_id = Geom::Transformation.new
	
	#Default value
	@height = 32
	@top_pos = true
	@message = nil
	@main_tooltip = ""
	@long_click_duration = 0.7
	persistence_restore

	#Initialize the structures
	init_struct

	#Parsing the arguments
	args.each do |arg|	
		arg.each { |key, value|  parse_palette_args(key, value) } if arg.class == Hash
	end
	
	#initialization of lists and hash tables	
	@lst_symbs = nil
	@lst_buttons = []
	@hsh_buttons = {}
	@button_blasons = []
	@post_drawing_buttons = []
	@dropdown_cur = nil
	
	@hsh_sides = {}
	@lst_sides = []
	@default_sidepal = :__side__pal
	
	@hsh_floats = {}
	@lst_floats = []
	@auto_hide_floats = false
		
	@hsh_visible_families = {}
	@current_family = []
	
	#Initialization for Bandeau
	@lst_bandeau = []
	@hgt_bandeau = 0
	@dec_bandeau = 4
	
	#Geometric transformation initialization
	@height2 = @height / 2
	@widmax_tip = 400 unless @widmax_tip
	@wid_ctrl = @height
	@lrectnum = [0, 0, 1, 1]
	@tra_basis = [1, 0, 0, 0] + [0, -1, 0, 0] + [0, 0, 1, 0] + [0, 0, 0, 1]
	@computed = false	
	@ptxy = Geom::Point3d.new 0, 0, 0
	@origin = Geom::Point3d.new 0, @height, 0
	
	#Color initialization
	@bkcolor_normal = 'lightgrey'
	@bkcolor_bandeau = @bkcolor_normal
	@bkcolor_hilight = 'lightgrey'
	@bkcolor_selected = 'oldlace'
	@bkcolor_title1_u = 'silver'
	@bkcolor_title1_d = 'gainsboro'
	@bkcolor_down = 'oldlace'
	@bkcolor_blason = 'khaki'
	@frcolor_blason = 'gold'
	@frcolor_white = 'white'
	@frcolor_gray = 'gray'
	@frcolor_dark = 'darkgray'
	@frcolor_float = 'darkgray'
	@bkcolor_float = 'khaki'
	@bdcolor_float = 'silver'
	@titcolor_float = 'khaki'
	@hibkcolor_float = 'silver'
	@hifrcolor_float = 'red'
	
	@id_cursor_hand = MYPLUGIN.create_cursor "Cursor_Hand", 9, 2	
	@id_cursor_arrow = MYPLUGIN.create_cursor "Cursor_ArrowPal", 2, 2	
	@id_cursor_move = MYPLUGIN.create_cursor "Cursor_Move_16", 7, 7	
	
	@tooltip = ""
	@nb_sepa = 0
	
	#Open GL instance for drawing buttons
	@ogl = Traductor::OpenGL_6.new
	@standard_draw_proc = @ogl.method "draw_proc"
	
	#Create the standard buttons
	@button_shrink = create_button :_bt_shrink, { :tooltip => T6[:Palette_TIP_Shrink], :width => @height2, :rank => 1 }
	@button_pos = create_button :_bt_pos, { :tooltip => T6[:Palette_TIP_Pos], :width => @height2 }
	@button_msg_visi = create_button :_bt_msg_visi, { :tooltip => T6[:Palette_TIP_MsgVisible], :width => @height2, 
	                                 :rank => 1, :text => '--' }
	@button_sideleft = create_button :_bt_sideleft, { :tooltip => T6[:Palette_TIP_SideLeft], :width => @height2 }
	@button_tip = create_button :_bt_tip, { :tooltip => T6[:Palette_TIP_TipZone], :passive => true, :justif => 'LT' }
	@button_message = create_button :_bt_message, { :tooltip => T6[:Palette_TIP_MsgZone], :passive => true, 
	                                                :height => 16, :hidden => true, :justif => 'LT' }
	adjust_draw_standard_button
	
	if @no_main
		@button_shrink.hidden = true
		@button_pos.hidden = true
		@button_msg_visi.hidden = true
		@button_sideleft.hidden = true
		@button_tip.hidden = true
		@button_message.hidden = true
	end
end

#INIT: Initialize the structures
def init_struct
	lpat = [:draw_refresh, :draw_extra_proc, 
		    :action_proc, :value_proc, :grayed_proc, :hidden_proc, :tip_proc, :text_proc, :long_click_proc, :double_click_proc,
            :row, :sepa, :width, :height, :hidden, :bk_color, :hi_color, :rank, :tooltip, :text, :sel_text, :passive,
	        :main_color, :frame_color, :type, :edge_prop, :stipple, :default_value, :blason, :fkey, :draw_scale, :justif,
		    :value, :radio, :input, :floating, :family, :tab_style, :zone_proc, :cursor_proc, :post_draw_proc,
		    :dropdown_pal, :hover_proc, :frame_hilight, :no_hilight, :bandeau, :draw_style, :border_color,
			:table_button, :table_cell, :table]

	#Patterns for button parameter settings
	@std_pats = lpat.collect { |symb| Regexp.new("\\A" + symb.to_s, Regexp::IGNORECASE) }
	
	#Structure for Buttons
	@struct_Palette_Button6 = Struct.new :symb, :state, :tr, :tr1, :tr2, :tr_tx, :hidsepa, :composite, :bsuite,
                                         :quad, :quad1, :quad2, :buttons, :draw_proc, :draw_code, :instructions,
							             :sel_instructions, :gray_instructions, :text_instructions, :text_sel_instructions,
                                         :xleft, :ybot, :computed, :parent, :value, :multi, :grayed, :selected, 
							             :sidepal, :numval, :bfloating, :text_previous, :text_sel_previous, 
										 :buttons_info_pos,
							             *lpat

	#Structures for Floating palettes
	@struct_Palette_Floating6 = Struct.new :symb, :title, :hidden, :computed, :buttons, :but_rows, :cur_row, :quad_out, :quad_in, 
                                           :quad_bd, :quad_tit, :pt_text, :line1, :line2, :line_tit, :xpos, :ypos, :empty,
							               :hidden_proc, :xpos0, :ypos0, :auto_move, :tretina, :follow_mouse, 
							               :bk_color, :fr_color, :tit_color, :bd_color, :justif_title, :dropdown_proc,
							               :invoked, :no_manual_move, :docking, :width, :height, :lysize, :lbmap

end

#INIT: Initialize the Retina factor
def retina_init
	@retina_factor = 1.5
end

#INIT: Assign the individual properties for the palette
def parse_palette_args(key, value)
	skey = key.to_s
	case skey
	when /top_pos/i
		@top_pos = value
	when /shrinked/i
		@shrinked = value
	when /height/i
		@height = value
	when /notify_change_position_proc/
		@notify_change_position_proc = value
	when /proc/i
		@palette_proc = value
	when /width_message_min/i
		@widmin_tip = value
	when /width_message/i
		@widmax_tip = value
	when /long_click_duration/i
		@long_click_duration = value
	when /key_registry/i
		@key_registry = value.to_s
	when /no_message/i
		@no_message = value
	when /no_main/i
		@no_main = value
	end	
end

#Parse the properties of a button
def parse_button_args(button, key, value)
	skey = key.to_s
	param = @std_pats.find { |pat| skey =~ pat }
	if param
		eval "button.#{param.source[2..-1]} = value"
		button.computed = false
		return
	end
	
	#Special treatment
	case skey
	when /parent/i
		button_set_parent button, value
	when /draw_proc/i
		lvalue = (value.class == Array) ? value : [value]
		lvalue.each do |val|
			if val.class == Symbol
				button.draw_code = [] unless button.draw_code
				button.draw_code.push val
			else	
				button.draw_proc = val
			end
		end	
	when /sidepal/i
		button.sidepal = (value.class == Symbol) ? value : @default_sidepal
	end
	button.computed = false	
end

#Return the height of the main palette in pixel
def height_main
	@height + 3
end

#Hide or show the mian palette
def set_no_main(flag)
	@no_main = flag
end

#Return the button structure from its symbol
def [](symb)
	@hsh_buttons[symb]
end

#Return the main tooltip of the palette
def get_main_tooltip
	@main_tooltip
end
	
#Get the button value state
def button_get_value(symb=nil)
	button = (symb) ? @hsh_buttons[symb] : @sel_button
	(button) ? button.value : nil
end

#Get the button value state
def button_is_grayed?(symb=nil)
	button = (symb) ? @hsh_buttons[symb] : @sel_button
	button.grayed
end

#Create a button structure
def create_button(symb, *args, &action_proc)
	button = @hsh_buttons[symb]
	button = @struct_Palette_Button6.new unless button
	button.symb = symb
	button.hidden = false
	button.grayed = false
	button.action_proc = action_proc
	button.family = @current_family
	button.rank = 0
	
	args.each do |arg|	
		arg.each { |key, value|  parse_button_args(button, key, value) } if arg.class == Hash
	end
	@hsh_buttons[symb] = button
	button
end

#Declare the parent of a button
def button_set_parent(button, sparent)
	parent = @hsh_buttons[sparent]
	parent = create_button sparent unless parent
	parent.buttons = [] unless parent.buttons
	parent.buttons.push button unless parent.buttons.include?(button)
	button.parent = parent
end

#Set the parameters for a button as a hash table of properties
def button_set_args(symb, *args)
	button = @hsh_buttons[symb]
	return unless button
	args.each do |arg|	
		arg.each { |key, value|  parse_button_args(button, key, value) } if arg.class == Hash
	end
	button.computed = false
	@computed = false
end

def button_set_draw_proc(symb, draw_proc, main_color=nil, frame_color=nil)
	button = @hsh_buttons[symb]
	return unless button
	parse_button_args button, :draw_proc, draw_proc
	parse_button_args button, :main_color, main_color if main_color
	parse_button_args button, :frame_color, frame_color if frame_color
	button_refresh_draw symb
end

def button_set_draw_standard_proc(symb, draw_code, main_color=nil, frame_color=nil)
	button = @hsh_buttons[symb]
	return unless button
	button.draw_code = draw_code
	parse_button_args button, :main_color, main_color if main_color
	parse_button_args button, :frame_color, frame_color if frame_color
	button_refresh_draw symb
end

def button_refresh_draw(symb)
	button = @hsh_buttons[symb]
	return unless button
	button.instructions = nil
	button.sel_instructions = nil
	button.gray_instructions = nil
	button.computed = false
end

def adjust_draw_standard_button
	button_set_draw_standard_proc :_bt_shrink, ((@shrinked) ? :square_P : :cross_NE2), "red"
	button_set_draw_standard_proc :_bt_pos, ((@top_pos) ? :triangle_SP : :triangle_NP), "green"
	button_set_draw_standard_proc :_bt_msg_visi, ((@message_visi) ? :square_1 : :square_P), "yellow"
	button_set_draw_standard_proc :_bt_sideleft, ((@side_left) ? :triangle_EP : :triangle_WP), "green"
end

def default_button_specs(*hargs)
	hsh = {}
	hargs.each do |harg|
		hsh.update harg if harg.class == Hash
	end
	@button_default_specs = hsh
end

#DECLARE: Declare a slider button
def declare_slider_button(button_symb, hslider_specs, *hargs)
	PaletteSliderButton.new self, button_symb, hslider_specs, *hargs
end

#Declare a button
def declare_button(symb, *args, &action_proc)
	button_exist = @hsh_buttons[symb]

	#Setting the button with default specs if any
	args = [@button_default_specs] + args if @button_default_specs

	button = create_button symb, *args, &action_proc
	button_input_manage button if button.input
	@computed = false
	@button_message.computed = false if @button_message
	
	#Button already in the palette
	return symb if button_exist
	
	#Adding the button in the palette
	if button.blason
		@button_blasons.push button
	elsif button.table_button	
		add_button_to_table(button)
	elsif button.sidepal
		add_button_to_side button
	elsif button.floating
		add_button_to_floating button
	else
		contribute_bandeau(button)
		@lst_buttons.push button unless button.parent
	end	

	#Associating the button to its table if applicable
	puts "button table = #{button.table}" if button.table.class == PaletteTable
	button.table.associate button if button.table.class == PaletteTable
	
	#Registration for post_drawing
	@post_drawing_buttons.push button if button.post_draw_proc
	
	symb
end

#Add a button to a table
def add_button_to_table(button)
	table_button = button.table_button
	table_button.buttons = [] unless table_button.buttons
	table_button.buttons.push button
	
	#Post processing
	button.floating = table_button.floating
	button.bfloating = table_button.bfloating
	
	#Returning the button
	button
end

#Declare a separator
def declare_separator
	s = '__sepa__' + @nb_sepa.to_s
	@nb_sepa += 1
	symb = s.intern
	declare_button symb, { :sepa => true, :passive => true, :height => @height, :width => @height / 3, 
	                       :draw_proc => :separator_V }
	symb
end

#Declare a separator for a side palette
def declare_separator_side(sidepal=nil, *args)
	s = '__sepa__' + @nb_sepa.to_s
	@nb_sepa += 1
	symb = s.intern
	declare_button symb, { :sidepal => sidepal, :passive => true, :height => @height / 3, :width => @height, 
	                       :draw_proc => :separator_H }, *args
	symb
end

#Declare a separator for a floating palette
def declare_separator_floating(symb_float, *args)
	s = '__sepa__' + @nb_sepa.to_s
	@nb_sepa += 1
	symb = s.intern
	declare_button symb, { :floating => symb_float, :sepa => true, :passive => true, :height => @height, :width => @height / 3, 
	                       :draw_proc => :separator_V }, *args
	symb
end

#Utility to form a sumbol from a symbol and text to append
def concat_symb(symb, text)
	s = symb.to_s + '___' + text
	s.intern
end

#Set the current family key for the subsequent button creation
def set_current_family(*family)
	@current_family = family
end

#Declare the visible families
def visible_family(*lst_families)
	@hsh_visible_families = {}
	lst_families.each { |family| @hsh_visible_families[family] = true }
	@computed = false
end

#----------------------------------------------------------------------------
# BANDEAU: Management of bandeau
#----------------------------------------------------------------------------

#BANDEAU: Declare a bandeau
def declare_bandeau(symb, *args, &action_proc)
	hsh = { :bandeau => true }
	declare_button(symb, hsh, *args, &action_proc)
	button = @hsh_buttons[symb]
	@lst_bandeau.push button 
	button.buttons = []
	@bandeau_cur = button
	@hgt_bandeau = 16 + @dec_bandeau

end

#BANDEAU: Reset the current bandeau
def no_bandeau
	@bandeau_cur = nil
end

#BANDEAU: Register the button in the current badeau if any 
def contribute_bandeau(button)
	@bandeau_cur.buttons.push button if @bandeau_cur
end

#BANDEAU: Compute the bandeau elements and Bandeau buttons 
def compute_bandeau
	return if @lst_bandeau.empty?
	
	#Computing the height of the bandeau
	@lst_bandeau.each do |button|	
		compute_button_bandeau button
	end
end

#BANDEAU: Compute a Bandeau button
def compute_button_bandeau(button)
	button.height = @hgt_bandeau - @dec_bandeau
	button_check_hidden?(button)
	button.computed = true
	return if button.hidden || button.buttons.empty?
	
	#Computing the position in X and width of the bandeau button
	xmin = xmax = nil
	pt = ORIGIN.clone
	button.buttons.each do |b|
		next if b.hidden || !b.quad || !b.tr 
		x1 = b.xleft
		next unless x1
		x2 = x1 + b.width
		xmin = x1 if !xmin || x1 < xmin
		xmax = x2 if !xmax || x2 > xmax		
	end
	return if xmax == xmin
	
	#Shift if of the palette is on the right side
	unless @side_left
		vpwidth = @view.vpwidth
		xmax2 = xmax
		xmax = vpwidth - xmin
		xmin = vpwidth - xmax2
	end
	
	#Computing the actual coodinates of the button
	button.width = xmax - xmin
	button_make_quads button
	xb = xmin
	yb = (@top_pos) ? @yref - @height - @dec_bandeau : @yref + @hgt_bandeau
	button_transfo button, xb, yb	
end

#----------------------------------------------------------------------------
# Special buttons: Edge properties
#----------------------------------------------------------------------------

#Declare a button set for edge property control
def declare_edge_prop(symb, rgshow, *args, &action_proc)
	declare_button(symb, { :type => 'multi' }, *args, &action_proc)
	button = @hsh_buttons[symb]
	hp = { :parent => symb, :width => 30, :main_color => button.main_color }
	button.edge_prop = 'PSMH' unless button.edge_prop
	list = []
	list.push ['P', T6[:T_DLG_EdgePlain], :edge_prop_plain] if button.edge_prop =~ /P/i
	list.push ['S', T6[:T_DLG_EdgeSoft], :edge_prop_soft] if button.edge_prop =~ /S/i
	list.push ['M', T6[:T_DLG_EdgeSmooth], :edge_prop_smooth] if button.edge_prop =~ /M/i
	list.push ['H', T6[:T_DLG_EdgeHidden], :edge_prop_hidden] if button.edge_prop =~ /H/i
	list.each do |ll|
		code, tip, sdraw = ll
		next unless !rgshow || code =~ rgshow
		declare_button concat_symb(symb, code), hp, { :draw_proc => sdraw, :value => code, :tooltip => tip }
	end	
	symb
end

#----------------------------------------------------------------------------
# Special buttons: Edge properties
#----------------------------------------------------------------------------

#Declare a button set for extended edge property control
# - get_proc and set_proc manipulates a Hash array of property: :smooth, :soft, :hidden, :cast_shadows
def declare_edge_prop_extended(prefix, symb, get_proc, set_proc, *hargs)
	#Plain and Diagonal
	hshb = { :width => 24, :height => 16 }
	value_proc = proc { a = get_proc.call(symb) ; a[:soft] == -1 && a[:smooth] == -1 && a[:hidden] == -1 && a[:cast_shadows] == 1 }
	action_proc = proc { a = get_proc.call(symb) ; a[:soft] = a[:smooth] = a[:hidden] = -1 ; a[:cast_shadows] = 1 ; set_proc.call symb, a }
	s = "#{prefix}__EPX_Plain".intern
	hsh = { :value_proc => value_proc, :draw_proc => :edge_prop_plain, :tooltip => T6[:T_DLG_EdgePlain], :rank => 1 }
	declare_button(s, hshb, hsh, { :rank => 1 }, *hargs, &action_proc)
	
	value_proc = proc { a = get_proc.call(symb) ; a[:soft] == 0 && a[:smooth] == 0 && a[:hidden] == 0 && a[:cast_shadows] == -1 }
	action_proc = proc { a = get_proc.call(symb) ; a[:soft] = a[:smooth] = a[:hidden] = 0 ; a[:cast_shadows] = -1 ; set_proc.call symb, a }
	s = "#{prefix}__EPX_Diagonal".intern
	hsh = { :value_proc => value_proc, :draw_proc => :edge_prop_diagonal, :tooltip => T6[:T_DLG_EdgeDiagonal] }
	declare_button(s, hshb, hsh, *hargs, &action_proc)

	declare_separator
	
	#Soft
	hshb = { :width => 32, :height => 16 }
	value_proc = proc { a = get_proc.call(symb) ; a[:soft] == 1 }
	action_proc = proc { a = get_proc.call(symb) ; a[:soft] = (a[:soft] == 1) ? 0 : 1 ; set_proc.call symb, a }
	s = "#{prefix}__EPX_Soft".intern
	hsh = { :value_proc => value_proc, :draw_proc => :edge_prop_soft, :tooltip => T6[:T_DLG_EdgeSoft], :rank => 1 }
	declare_button(s, hshb, hsh, *hargs, &action_proc)
	
	value_proc = proc { a = get_proc.call(symb) ; a[:soft] == -1 }
	action_proc = proc { a = get_proc.call(symb) ; a[:soft] = (a[:soft] == -1) ? 0 : -1 ; set_proc.call symb, a }
	s = "#{prefix}__EPX_SoftNo".intern
	hsh = { :value_proc => value_proc, :draw_proc => [:edge_prop_soft, :std_negation], :tooltip => T6[:T_DLG_EdgeSoft] }
	declare_button(s, hshb, hsh, *hargs, &action_proc)

	#Smooth
	value_proc = proc { a = get_proc.call(symb) ; a[:smooth] == 1 }
	action_proc = proc { a = get_proc.call(symb) ; a[:smooth] = (a[:smooth] == 1) ? 0 : 1 ; set_proc.call symb, a }
	s = "#{prefix}__EPX_Smooth".intern
	hsh = { :value_proc => value_proc, :draw_proc => :edge_prop_smooth, :tooltip => T6[:T_DLG_EdgeSmooth], :rank => 1 }
	declare_button(s, hshb, hsh, *hargs, &action_proc)
	
	value_proc = proc { a = get_proc.call(symb) ; a[:smooth] == -1 }
	action_proc = proc { a = get_proc.call(symb) ; a[:smooth] = (a[:smooth] == -1) ? 0 : -1 ; set_proc.call symb, a }
	s = "#{prefix}__EPX_SmoothNo".intern
	hsh = { :value_proc => value_proc, :draw_proc => [:edge_prop_smooth, :std_negation], :tooltip => T6[:T_DLG_EdgeSmooth] }
	declare_button(s, hshb, hsh, *hargs, &action_proc)
	
	#Hidden
	value_proc = proc { a = get_proc.call(symb) ; a[:hidden] == 1 }
	action_proc = proc { a = get_proc.call(symb) ; a[:hidden] = (a[:hidden] == 1) ? 0 : 1 ; set_proc.call symb, a }
	s = "#{prefix}__EPX_Hidden".intern
	hsh = { :value_proc => value_proc, :draw_proc => :edge_prop_hidden, :tooltip => T6[:T_DLG_EdgeHidden], :rank => 1 }
	declare_button(s, hshb, hsh, *hargs, &action_proc)
	
	value_proc = proc { a = get_proc.call(symb) ; a[:hidden] == -1 }
	action_proc = proc { a = get_proc.call(symb) ; a[:hidden] = (a[:hidden] == -1) ? 0 : -1 ; set_proc.call symb, a }
	s = "#{prefix}__EPX_HiddenNo".intern
	hsh = { :value_proc => value_proc, :draw_proc => [:edge_prop_hidden, :std_negation], :tooltip => T6[:T_DLG_EdgeHidden] }
	declare_button(s, hshb, hsh, *hargs, &action_proc)

	#Cast Shadows
	value_proc = proc { a = get_proc.call(symb) ; a[:cast_shadows] == 1 }
	action_proc = proc { a = get_proc.call(symb) ; a[:cast_shadows] = (a[:cast_shadows] == 1) ? 0 : 1 ; set_proc.call symb, a }
	s = "#{prefix}__EPX_CastShadows".intern
	hsh = { :value_proc => value_proc, :draw_proc => :edge_prop_cast_shadows, :tooltip => T6[:T_DLG_EdgeCastShadows], :rank => 1 }
	declare_button(s, hshb, hsh, *hargs, &action_proc)
	
	value_proc = proc { a = get_proc.call(symb) ; a[:cast_shadows] == -1 }
	action_proc = proc { a = get_proc.call(symb) ; a[:cast_shadows] = (a[:cast_shadows] == -1) ? 0 : -1 ; set_proc.call symb, a }
	s = "#{prefix}__EPX_CastShadowsNo".intern
	hsh = { :value_proc => value_proc, :draw_proc => [:edge_prop_cast_shadows, :std_negation], :tooltip => T6[:T_DLG_EdgeCastShadows] }
	declare_button(s, hshb, hsh, *hargs, &action_proc)
	
	symb
end

#----------------------------------------------------------------------------
# Special buttons: Stipple properties
#----------------------------------------------------------------------------

#Declare a button set for edge property control
def declare_stipple(symb, *args, &action_proc)
	declare_button(symb, { :type => 'multi', :radio => true }, *args, &action_proc)
	button = @hsh_buttons[symb]
	hp = { :parent => symb, :width => 25, :main_color => button.main_color, :hi_color => button.hi_color }
	stipple = button.stipple
	stipple = 'DUPA' unless stipple
	list = []
	list.push ['C', T6[:T_TIP_Stipple_Continuous]] if stipple =~ /C/i
	list.push ['D', T6[:T_TIP_Stipple_Dash]] if stipple =~ /D/i
	list.push ['U', T6[:T_TIP_Stipple_Underscore]] if stipple =~ /U/i
	list.push ['P', T6[:T_TIP_Stipple_Dot]] if stipple =~ /P/i
	list.push ['A', T6[:T_TIP_Stipple_Alternate]] if stipple =~ /A/i
	list.each do |ll|
		code, tip = ll
		sdraw = "line_#{code}2".intern
		declare_button concat_symb(symb, code), hp, { :draw_proc => sdraw, :value => code, :tooltip => tip }
	end	
	symb
end

#----------------------------------------------------------------------------
# Special buttons: Input field
#----------------------------------------------------------------------------

def button_input_manage(button)
	input = button.input
	button.text_proc = proc { input.compute_show_text nil, true} unless button.text_proc
	button.action_proc = proc { input.dialog_ask } unless button.action_proc
	button_input_increment(button) if input.vincr || input.vbound_proc
end

def button_input_increment(button)
	#dec = 14
	dec = 13
	button.width = 80 unless button.width
	button.width -= 2 * dec
	button.composite = :plus_minus
	ss = button.symb.to_s
	hsh = { :height => button.height, :width => dec, :rank => button.rank, :row => button.row, 
	        :bk_color => button.bk_color, :main_color => 'gray' }
	
	symb = (ss + '___plus').intern	
	gray_proc = proc { button.input.reached_max? }
	hshplus = { :draw_proc => :triangle_EP, :grayed_proc => gray_proc, :tooltip => T6[:T_TIP_Increment] }
	bplus = create_button(symb, hsh, hshplus) { button.input.increment }
	button_set_parent bplus, button.symb
	bplus.bfloating = button.bfloating
	
	symb = (ss + '___minus').intern
	gray_proc = proc { button.input.reached_min? }
	hshminus = { :draw_proc => :triangle_WP, :grayed_proc => gray_proc, :tooltip => T6[:T_TIP_Decrement] }
	bminus = create_button(symb, hsh, hshminus) { button.input.decrement }
	button_set_parent bminus, button.symb
	bminus.bfloating = button.bfloating
	#bminus.family = button.family
	
	button.bsuite = [bminus, button, bplus]
end

#----------------------------------------------------------------------------
# Historical button group
#----------------------------------------------------------------------------

#Declare a button
def declare_historical(symb, *args, &action_proc)
	#Parsing the args specific to the Historical button group
	grayed_proc = nil
	tip_proc = nil
	compact = false
	title = 'Edition'
	floating = nil

	args.each do |arg|	
		next unless arg.class == Hash
		arg.each do |key, value| 
			skey = key.to_s
			case skey
			when /grayed_proc/i
				grayed_proc = value
			when /tip_proc/i
				tip_proc = value
			when /compact/i	
				compact = value
			when /title/i
				title = value
			when /floating/i
				floating = value
			when /key_ctrl/
				@key_ctrl = value
			when /key_shift/
				@key_shift = value
			end	
		end
	end
	
	symb_left = symbol_arrow :left, @key_ctrl, @key_shift
	symb_right = symbol_arrow :right, @key_ctrl, @key_shift
	symb_up = symbol_arrow :up, @key_ctrl, @key_shift
	symb_down = symbol_arrow :down, @key_ctrl, @key_shift
	
	#List of buttons
	lsbut = [[:undo, 1, :std_undo, 'black', T6[:T_TIP_EditionUndo], [:escape, symb_left]],
			 [:clear, 0, :cross_NE3, 'red', T6[:T_TIP_EditionReset], [symb_down]],
	         [:redo, 1, :std_redo, 'black', T6[:T_TIP_EditionRedo], [symb_right]],
			 [:restore, 0, :std_restore, 'green', T6[:T_TIP_EditionRestore], [symb_up]]]
			 
	#Creating the buttons
	libut = (compact) ? [0, 1, 2, 3] : [1, 0, 2, 3]
	wid = hgt = 16
	if !compact && title
		hsh = { :passive => true, :text => title, :height => hgt, :width => 4 * wid, :rank => 1 }
		h = args + [hsh]
		declare_button symb, *h
	end	
	txt_ctrl = ''
	txt_ctrl = T6[:T_KEY_Ctrl] + ' ' if @key_ctrl
	txt_ctrl = T6[:T_KEY_Shift] + ' ' if @key_shift
	libut.each do |i|		
		info = lsbut[i]
		ss = symb.to_s + info[0].to_s
		rank = (compact) ? info[1] : 0
		hsh = { :height => hgt, :width => wid, :rank => rank,
		        :draw_proc => info[2], :frame_color => info[3] }
		hsh[:grayed_proc] = proc { grayed_proc.call info[0] } if grayed_proc
		hsh[:tip_proc] = proc { tip_proc.call info[0] } if tip_proc
		hsh[:action_proc] = proc { action_proc.call info[0] } if action_proc
		hsh[:tooltip] = Traductor.encode_tip info[4], info[5]
		h = args + [hsh]
		declare_button ss.intern, *h
	end	
end

def symbol_arrow(symb, ctrl, shift)
	if ctrl
		s = "ctrl_arrow_#{symb}"
	elsif shift
		s = "shift_arrow_#{symb}"
	else	
		s = "arrow_#{symb}"
	end
	s.intern	
end

#----------------------------------------------------------------------------
# Side Palette
#----------------------------------------------------------------------------

#Register a button in a side palette
def add_button_to_side(button)
	sidepal = button.sidepal
	lst = @hsh_sides[sidepal.to_s]
	unless lst
		lst = @hsh_sides[sidepal.to_s] = []
		@lst_sides.push lst
	end
	lst.push button
end

#Compute the side palettes
def compute_side_palettes
	#Computing maximum width of buttons
	@total_width_side = 0
	vpwidth = @view.vpwidth
	vpheight = @view.vpheight
	lwidmax = []
	totwid = 0
	@lst_sides.each_with_index do |lst, i|
		lst.each { |button|	lwidmax[i] = button.width unless button.hidden }
		totwid += lwidmax[i]
	end	
	
	#No visible buttons in the side palette
	@visible_sidepal = (totwid > 0)
	return unless @visible_sidepal
	
	#Dimensions and starting points
	@total_width_side = totwid	
	hpal = @total_height + 1
	xbeg = 0
	ybeg = (@top_pos) ? hpal-1 : vpheight - hpal
	
	#Computing the position of the buttons
	lhgtmax = []
	xpos = xbeg
	@lst_sides.each_with_index do |lst, i|
		lhgtmax[i] = 0
		next if lwidmax[i] == 0
		ypos = ybeg
		lst.each do |button|
			next if button.hidden
			ypos += button.height if @top_pos && button.rank == 0
			h = (@top_pos || button.rank == 0) ? 0 : button.height
			button_transfo button, xpos + button.rank * button.width, ypos + h
			ypos -= button.height if !@top_pos && button.rank == 0
			lhgtmax[i] += button.height if button.rank == 0
		end
		xpos += lwidmax[i] + 1
	end
	
	#Computing the quad for the side palettes
	@total_height_side = lhgtmax.max + 4
	@xbeg_side = xbeg
	@ybeg_side = ybeg
	xmin = xbeg
	xmax = xmin + @total_width_side
	ymin = ybeg
	ymax = ymin + @total_height_side * ((@top_pos) ? 1 : -1)
	@quad_sidepal = []
	@quad_sidepal.push Geom::Point3d.new(xmin, ymin)
	@quad_sidepal.push Geom::Point3d.new(xmax, ymin)
	@quad_sidepal.push Geom::Point3d.new(xmax, ymax)
	@quad_sidepal.push Geom::Point3d.new(xmin, ymax)
	
end

#----------------------------------------------------------------------------
# Floating palette
#----------------------------------------------------------------------------

#Declare a floating palette
def declare_floating(symb, *args)
	#Creating the data structure if not already created
	floating = @hsh_floats[symb]
	unless floating
		floating = @struct_Palette_Floating6.new
		floating.symb = symb
		floating.buttons = []
		floating.but_rows = []
		floating.cur_row = 0
		floating.bk_color = @bkcolor_float
		floating.fr_color = @frcolor_float
		floating.tit_color = @titcolor_float
		floating.bd_color = @bdcolor_float
		floating_initial_position floating
		@hsh_floats[symb] = floating
		@lst_floats.push floating
	end
	
	#parsing the arguments
	args.each do |arg|	
		arg.each { |key, value|  parse_floating_args(floating, key, value) } if arg.class == Hash
	end
	symb
end

#Set the auto_hide mode for floating palettes
def auto_hide_floating(on=false)
	@auto_hide_floats = on
	@hidden_floating = nil
end

def show_hide_floating(symb_float, show=true)
	floating = @hsh_floats[symb_float]
	return false if floating == nil || floating.hidden == !show
	floating.hidden = !show
	true
end

#Calculate or retrieve the initial position of the floating palette
def floating_initial_position(floating)
	xpos = ypos = nil
	if @key_registry
		ss = 'Floating_Palette_' + floating.symb.to_s
		sxypos = Sketchup.read_default @key_registry, ss
		if sxypos
			l = sxypos.split ';'
			xpos = Traductor.string_to_integer l[0]
			ypos = Traductor.string_to_integer l[1]
		end	
	end	
	floating.xpos = (xpos) ? xpos : @view.vpwidth / 2
	floating.ypos = (ypos) ? ypos : 60
	floating.xpos0 = floating.xpos
	floating.ypos0 = floating.ypos
end

#Store the floating palette position into the Sketchup registry
def floating_store_registry(floating)
	return unless @key_registry
	ss = 'Floating_Palette_' + floating.symb.to_s
	sxypos = floating.xpos.round.to_s + ';' + floating.ypos.round.to_s
	Sketchup.write_default @key_registry, ss, sxypos
end

#Get the current floating position
def floating_get_position(symb_float)
	floating = @hsh_floats[symb_float]
	return nil unless floating
	[floating.xpos0, floating.ypos0]
end

#Get the current floating position
def floating_set_position(symb_float, xpos, ypos)
	floating = @hsh_floats[symb_float]
	return nil unless floating
	floating.xpos0 = xpos
	floating.ypos0 = ypos
end

#Check if a floating palette should be hidden
def floating_is_hidden?(floating)
	return true if floating.hidden || floating.empty || floating == @hidden_floating
	return true if floating.hidden_proc.call if floating.hidden_proc
	return true if floating.dropdown_proc && floating != @dropdown_cur
	false
end
	
#Assign the individual property for the floating palette
def parse_floating_args(floating, key, value)
	skey = key.to_s
	case skey
	when /justif_title/i
		floating.justif_title = value
	when /title/i
		floating.title = value
	when /docking/i
		floating.docking = value
	when /xpos/i
		floating.xpos = value
	when /ypos/i
		floating.ypos = value
	when /hidden_proc/i
		floating.hidden_proc = value
	when /hidden/i
		floating.hidden = value
	when /auto_move/i
		floating.auto_move = value
	when /follow_mouse/i
		floating.follow_mouse = value
	when /bk_color/i
		floating.bk_color = value
	when /fr_color/i
		floating.fr_color = value
	when /tit_color/i
		floating.tit_color = value
	when /bd_color/i
		floating.bd_color = value
	when /dropdown_proc/i
		floating.dropdown_proc = value
	when /no_manual_move/i
		floating.no_manual_move = value
	end	
end

#Change the title of a floating palette
def floating_set_title(symb_float, title)
	floating = @hsh_floats[symb_float]
	floating.title = title if floating
end

#Change the title of a floating palette
def floating_set_follow_mouse(symb_float, follow_mouse)
	floating = @hsh_floats[symb_float]
	floating.follow_mouse = follow_mouse if floating
end

#Register a button in a side palette
def add_button_to_floating(button)
	#Determining the Floating palette object
	symb_float = button.floating
	floating = @hsh_floats[symb_float]
	unless floating
		declare_floating symb_float
		floating = @hsh_floats[symb_float]
	end
	button.bfloating = floating
	button.bsuite.each { |b| b.bfloating = floating } if button.bsuite
	
	#Adding the buttons to the floating palette
	unless button.parent
		floating.buttons.push button
		row = (button.row) ? button.row : floating.cur_row
		floating.cur_row = row if row
		floating.but_rows[row] = [] unless floating.but_rows[row]
		floating.but_rows[row].push button
	end	
end

#FLOATING: Compute the Floating palette - Position of buttons and palette
def compute_floating_palette(floating)
	#Dimension of the palette
	lxsize = []
	lysize = []
	hvsepa = 2
	sxmax = 0
	symax = 0
	lbmap = []
	
	floating.buttons.each { |button| button_check_hidden?(button) }
	
	floating.but_rows.each_with_index do |but_row, irow|
		no_vsepa = false
		x = sy = 0
		h = hvsepa
		cursepa = true
		xrank = 0
		curank = nil
		syprev = @height

		#All buttons are hidden
		unless but_row.find { |button| !button_check_hidden?(button) }
			lysize[irow] = 0
			next
		end	
		
		#Finding the smallest rank displayed
		begrank = nil
		but_row.each do |button|
			next if button.hidden
			begrank = button.rank if !begrank || button.rank < begrank
		end	

		but_row.each do |button|
			next if button.hidden
			no_vsepa = true if button.tab_style
			if button.sepa
				if cursepa || button == but_row.last
					button.hidsepa = true
					next
				else
					button.hidsepa = false
					cursepa = true
				end
			else
				cursepa = false
			end	
			rank = button.rank
			if rank != curank #&& button.type !~ /multi/i
				h += button.height
				xrank = 0
			end	
			
			if button.type =~ /multi/i
				lbmap.push [button, irow, x + xrank, symax + h]
				xb = x
				h2 = 0
				lsbut = []
				button.buttons.each { |b| lsbut += (b.bsuite) ? b.bsuite : [b] }
				lsbut.each do |b|
					next if b.hidden
					h2 = b.height if h2 < b.height
					lbmap.push [b, irow, xb, symax + h + h2]
					xb += b.width
					no_vsepa = true if b.tab_style
				end
				x += button.width
				sy = h + h2 if sy < h
				h -= button.height
			elsif button.bsuite
				button.bsuite.each do |b|
					lbmap.push [b, irow, x + xrank, symax + h]
					x += b.width
					no_vsepa = true if b.tab_style
				end	
			else	
				lbmap.push [button, irow, x + xrank, symax + h]
				x += button.width if button.rank == begrank
			end	
			
			sy = h if sy < h
			if rank == begrank
				syprev = h - hvsepa
				h = hvsepa
				curank = nil
			else
				curank = rank
				xrank += button.width
			end	
		end	
		sxmax = x if x > sxmax
		if no_vsepa
			lysize[irow] = -(sy + ((irow == 0) ? 0 : 3))
			symax += sy + 2
		else	
			lysize[irow] = sy + hvsepa + ((irow == 0) ? 0 : 2)
			symax += sy + hvsepa + 2
		end	
	end
	
	#No buttons to display
	if lbmap.empty?
		floating.empty = true
		return
	end	
	floating.empty = false
	
	#Position of the palette
	delta = 5
	
	topmargin = delta
	if floating.title
		topmargin += 16
		wt = 6 * G6.text_true_length(floating.title)
		sxmax = wt if wt > sxmax
	end	
	w = sxmax + 2 * delta
	h = symax + topmargin + delta - 2
	floating.width = w
	floating.height = h
	floating.lysize = lysize
	floating.lbmap = lbmap
	
	#Computing the position of the floating palette
	floating_compute_position(floating) unless floating.docking
end

#FLOATING: Compute the position of the Floating palette and its buttons
def floating_compute_position(floating)
	w = floating.width
	h = floating.height
	lysize = floating.lysize
	lbmap = floating.lbmap
	
	#Dimension of the palette
	delta = 5	
	topmargin = delta
	topmargin += 16 if floating.title
	sxmax = w - 2 * delta
	symax = h - topmargin - delta + 2
	
	#Making sure the floating palette remains within bounds
	xpos = floating.xpos0
	ypos = floating.ypos0
	xpos, ypos = floating_check_overlap(xpos, ypos, w, h) unless floating.docking
	floating.xpos = xpos
	floating.ypos = ypos
	
	#Retina transformation
	origin = Geom::Point3d.new xpos, ypos, 0
	tretina = floating.tretina = Geom::Transformation.scaling(origin, @retina_factor)
	
	#Quads for the palette
	floating.quad_out = floating_make_quad tretina, xpos, ypos, w, h
	floating.quad_in = floating_make_quad tretina, xpos + delta, ypos + topmargin, sxmax, symax - 2
	floating.quad_bd = floating_make_quad tretina, xpos + delta, ypos + topmargin, sxmax + 1, symax - 2
	
	#Adjustment for drop-down floating palette when in bottom position
	drop_rev = !@top_pos && @dropdown_cur && @dropdown_cur == floating
	hgt_in = symax - 2

	#Computing the row separators and title positions
	ytop = ypos + topmargin 
	yline = 0 
	xbeg = xpos + delta
	xend = xpos + w - delta
	floating.line1 = []
	floating.line2 = []
	for i in 0..lysize.length-2
		next if lysize[i] < 0
		yline += lysize[i].abs
		if drop_rev
			y1 = ytop + hgt_in - yline - 1
			y2 = y1 + 1
		else
			y1 = ytop + yline
			y2 = y1 - 1
		end
		floating.line1.push tretina * Geom::Point3d.new(xbeg, y1, 0), tretina * Geom::Point3d.new(xend, y1 , 0)
		floating.line2.push tretina * Geom::Point3d.new(xbeg, y2 , 0), tretina * Geom::Point3d.new(xend, y2 , 0)
	end
	if floating.title
		case floating.justif_title
		when /M/
			wtx, = G6.simple_text_size(floating.title)
			xbeg_tx = (xbeg + xend - wtx) / 2
		when /R/
			wtx, = G6.simple_text_size(floating.title)
			xbeg_tx = xend - wtx
		else
			xbeg_tx = xbeg
		end
		yltext = ypos + 18
		floating.pt_text = tretina * Geom::Point3d.new(xbeg_tx, ypos, 0)
		floating.line_tit = [tretina * Geom::Point3d.new(xpos, yltext, 0), tretina * Geom::Point3d.new(xpos + w, yltext , 0)]
		floating.quad_tit = floating_make_quad tretina, xpos, ypos, w, 18
	end	
		
	#Computing the final position of the buttons
	decx = xpos + delta
	decy = ypos + topmargin
	lbmap.each do |ll|
		button, irow, x, y = ll
		y = hgt_in - y + button.height if drop_rev
		button_transfo button, x + decx, y + decy, true
		if button.table
			puts "FLOATING SUB PARENTS"
			table_position_buttons(button, x + decx, y + decy)
		end
	end	
end

#Check the position of the floating palette based on viewport and other existing palettes
def floating_check_overlap(xpos, ypos, w, h)
	vpwidth = @view.vpwidth
	vpheight = @view.vpheight

	xpos = vpwidth - w - 1 if xpos + w > vpwidth
	xpos = 0 if xpos < 0
	ypos = vpheight - h - 1 if ypos + h > vpheight
	ypos = 0 if ypos < 0

	[xpos, ypos]
end

#generic routine to build a quad
def floating_make_quad(tretina, x, y, width, height)
	pt1 = Geom::Point3d.new x, y, 0
	pt2 = Geom::Point3d.new x + width, y, 0
	pt3 = Geom::Point3d.new x + width, y + height, 0
	pt4 = Geom::Point3d.new x, y + height, 0
	[pt1, pt2, pt3, pt4].collect { |pt| tretina * pt }
end

#Draw the floating palette background and title
def draw_floating_background(view, floating)
	return if floating_is_hidden?(floating)

	if floating == @mouse_floating && @origin_floating
		bkcolor = @hibkcolor_float
		frcolor = @hifrcolor_float
		wid = 2
	else
		bkcolor = floating.bk_color
		frcolor = floating.fr_color
		wid = 1
	end
	stipple = (@origin_floating) ? '-' : ''
	
	view.drawing_color = bkcolor
	view.line_width = wid
	view.line_stipple = stipple
	view.draw2d GL_QUADS, floating.quad_out
	if floating.title
		view.drawing_color = floating.tit_color
		view.draw2d GL_QUADS, floating.quad_tit
	end	
	view.drawing_color = @bkcolor_normal
	view.draw2d GL_QUADS, floating.quad_in
	view.drawing_color = frcolor
	view.draw2d GL_LINE_LOOP, floating.quad_out
	view.drawing_color = floating.bd_color
	view.line_width = 1
	view.draw2d GL_LINE_LOOP, floating.quad_bd
	if floating.title
		view.line_stipple = ''
		G6.view_draw_text view, floating.pt_text, floating.title
		view.drawing_color = @frcolor_dark
		view.draw2d GL_LINES, floating.line_tit
	end	
	view.drawing_color = @frcolor_white
	view.draw2d GL_LINES, floating.line1 unless floating.line1.empty?
	view.drawing_color = @frcolor_dark
	view.draw2d GL_LINES, floating.line2 unless floating.line2.empty?	
end

#Locate the floating frame where the mouse is
def locate_floating_frame(ptxy)
	@lst_floats.each do |floating|
		next if floating_is_hidden?(floating)
		next if floating.dropdown_proc || floating.no_manual_move || !floating.quad_out || floating.docking
		if Geom.point_in_polygon_2D(ptxy, floating.quad_out, true)
			if @mouse_floating != floating
				@mouse_floating_frame = !Geom.point_in_polygon_2D(ptxy, floating.quad_in, false)
				@mouse_floating = floating
				@view.invalidate
			end	
			return true
		end	
	end	
	@mouse_floating = nil
	false
end

#Locate the floating frame where the mouse is
def test_auto_hide_floating(x, y)
	return false unless @auto_hide_floats
	ptxy = Geom::Point3d.new x, y, 0
	@hidden_floating = nil
	@lst_floats.each do |floating|
		next if floating_is_hidden?(floating)
		if Geom.point_in_polygon_2D(ptxy, floating.quad_out, true)
			@hidden_floating = floating
			return true
		end	
	end	
	false
end

#Move a palette to a new origin
def move_floating(floating, x, y)
	return if floating.xpos == x && floating.ypos == y
	floating.xpos = floating.xpos0 = x
	floating.ypos = floating.ypos0 = y
	compute_floating_palette floating
	@view.invalidate
end

#Move the floating palette to their automatic position
def automove_floating(x, y)
	@lst_floats.each do |floating|
		type, dx, dy = floating.auto_move
		next unless type
		dx = 10 unless dx
		dy = 10 unless dy
		floating.xpos0 = x + dx
		floating.ypos0 = y + dy
		@computed = false
	end
end

#Move the floating palette to their automatic position
def follow_mouse_floating(x, y)
	@lst_floats.each do |floating|
		type, dx, dy = floating.follow_mouse
		next unless type
		dx = 10 unless dx
		dy = 10 unless dy
		qd = floating.quad_out
		w = qd[1].x - qd[0].x
		floating.xpos0 = (@view.vpwidth - x - dx < w) ? x - w - dx : x + dx
		floating.ypos0 = y + dy
		@computed = false
	end
end

#Move current highlighted palette
def move_current_floating(x, y)
	floating = @mouse_floating
	x += @origin_floating[0]
	y += @origin_floating[1]
	move_floating floating, x, y 
end

#Terminate the move of the floating palette
def end_move_floating
	floating_store_registry @mouse_floating if @mouse_floating
end

#----------------------------------------------------------------------------
# DOCKING: Management of Docking palettes
#----------------------------------------------------------------------------

#DOCKING: Main method to compute the position of docking palettes
def docking_compute
	return if @lst_floats.empty?
	
	#Computing the list of dockable palettes
	ls_floats = []
	@lst_floats.each do |floating|
		next if floating_is_hidden?(floating)
		type, fprev = docking_type(floating)
		next unless type
		ls_floats.push [type, floating]
	end
	return if ls_floats.empty?
	
	#Organizing the palette by row
	vpwidth = @view.vpwidth
	vpheight = @view.vpheight
	width_side = @total_width_side
	irow = 0
	x_row = @total_button_width
	y_row = 0
	hmax = nil
	ls_info = []
	ls_floats.each do |type, floating|
		w = floating.width
		h = floating.height
		type = :under if (x_row + w) > vpwidth
		case type
		when :under
			x = @total_width_side
			hmax = @total_height unless hmax
			y_row += hmax
			hmax = h
			x_row = x + w
		else
			x = x_row
			hmax = h if !hmax || h > hmax
			x_row += w			
		end
		ls_info.push [floating, x, y_row]
	end
	
	#Positioning the docked palettes
	ls_info.each do |floating, x, y|
		x = vpwidth - x - floating.width unless @side_left
		y = vpheight - y - floating.height unless @top_pos	
		floating.xpos0 = x
		floating.ypos0 = y
		floating_compute_position(floating)
	end	
end

#Return the type and ref for the docking parameter
def docking_type(floating)
	docking = floating.docking
	return nil unless docking
	(docking.class == Proc) ? docking.call : docking
end
	
#----------------------------------------------------------------------------
# Computation methods for button and Palette
#----------------------------------------------------------------------------

#Compute a button
def compute_button(button, forced=false)
	return if button.bandeau
	return if button.computed && !forced
	return compute_multi(button, forced) if button.type =~ /multi/i
	return if button.parent && button.parent.type =~ /multi/i
	compute_button_table(button, forced) if button.table
	button.rank = 0 unless button.rank
	button.width = button.height unless button.width
	button.width = @height unless button.width
	button.height = button.width unless button.height
	button_make_quads button
	button_check_hidden?(button)
	
	if button.composite && button.buttons
		button.buttons.each { |b| compute_button b, true }
	end
	
	button.computed = true
end

#Compute a Table button
def compute_button_table(button, forced=false)
	#Computing the inner buttons
	if button.buttons
		button.buttons.each { |b| compute_button b, forced }
	end
		
	table_compute_layout(button)
end

#Compute a multi-set of buttons
def compute_multi(button, forced=false)
	return if button.computed && !forced
	w = 0
	button.height = @height2 unless button.height
	lsbut = []
	button.buttons.each { |b| lsbut += ((b.bsuite) ? b.bsuite : [b]) }
	lsbut.each do |b|
		b.rank = 0
		b.height = button.height unless b.height
		b.computed = false
		b.floating = button.floating
		b.bfloating = button.bfloating
		compute_button b
		button_track_changes b
		w += b.width unless b.hidden
		button_make_quads b
		b.computed = true
	end
	button.width = w
	button.rank = 1
	button_make_quads button
	button.multi = true if button.type =~ /\Amulti\Z/i
	button.computed = true
end

def button_make_quads(button)
	button.quad = make_quad button.width, button.height
	button.quad1 = make_quad button.width, button.height, 1
	button.quad2 = make_quad button.width, button.height, 2
end

def check_multi_state(button)
	sval = button.value_proc.call if button.value_proc
	button.value = sval
	if button.radio
		lv = [sval]
	elsif sval.is_a? String
		lv = (sval) ? sval.split(';;') : nil
	else
		lv = sval
	end
	button.buttons.each do |b|
		if b.value != nil
			b.selected = (lv && !lv.include?(b.value)) ? false : true
		else
			b.selected = false
		end	
	end	
end

#Calculate the button tooltip
def button_compute_tooltip(button)
	tip = nil
	if button.tip_proc
		tip = button.tip_proc.call
	end
	unless tip
		if button.tooltip
			tip = button.tooltip
		elsif button.fkey && button.value_proc
			tip = button.fkey.build_text button.value_proc.call, true
		end
	end	
	(tip) ? tip : ""
end

def button_in_family?(button)
	return true if button.family.empty? || @hsh_visible_families.length == 0 
	button.family.each do |family|
		return true if @hsh_visible_families[family]
	end
	false
end

def compute_button_state(button)
	if button.multi 
		check_multi_state button
	elsif button.value_proc
		button.selected = button.value_proc.call
	end	
end

#Recompute the whole set of palettes
def compute_all
	#Compute the normal buttons
	@hsh_buttons.each { |symb, button| compute_button button, true} unless @computed
	
	#List of buttons
	lst_buttons = @lst_buttons.find_all { |button| button_in_family?(button) && !button.bandeau }	######
	
	#Checking button state
	@hsh_buttons.each do |key, button|
		compute_button_state button unless button.hidden || !button_in_family?(button)
	end
	
	#Already computed and viewport did not change size
	return if @computed && @vpwidth == @view.vpwidth && @vpheight == @view.vpheight
	
	#Reference viewport width and position for buttons
	@vpwidth = @view.vpwidth
	@vpheight = @view.vpheight
	vpwid = @view.vpwidth
	hgt_bandeau = (@shrinked) ? 0 : @hgt_bandeau
	@yref = (@top_pos) ? hgt_bandeau + @height + 1 : @view.vpheight + 1 - hgt_bandeau
	
	#Compute the width for the buttons
	widbut = 0
	lst_buttons.each do |button|
		widbut += button.width if (!(button.type =~ /multi/i) && button.rank == 0) || button.type =~ /multi/i
	end
	widblason = 0
	@button_blasons.each do |button| 
		button.width = @height
		widblason += button.width
		button_make_quads button		
	end	

	#Show or hide the tooltip  area
	unless @message_visi && !@no_main
		@button_tip.hidden = true
		wid_tip = 0
	else	
		@button_tip.hidden = false
		wid_tip = @widmax_tip
	end	
	
	#Button position starting reference
	@xstart_button = (@no_main) ? 0 : wid_tip + @wid_ctrl + widblason
	
	#Computing the position of the control buttons
	lbt = [@button_shrink, @button_pos, @button_msg_visi, @button_sideleft]
	lx = [0, 0, @height2, @height2]
	ly = [@height2, 0, @height2, 0]
	for i in 0..3
		b = lbt[i]
		x = lx[i]
		y = @yref - ((@top_pos) ? ly[i] : ly[3-i])
		button_transfo b, x, y
	end
	
	#Blason buttons
	w = 0
	@button_blasons.each do |button| 
		button_transfo button, @wid_ctrl + w, @yref
		w += button.width 
	end	
	
	#Computing the other buttons
	ls_buttons = []
	lst_transfo = []
	lst_buttons.each { |button| ls_buttons += (button.bsuite) ? button.bsuite : [button] }
	
	cursepa = false
	xpos = @xstart_button
	xrank = 0
	curank = nil
	ls_buttons.each do |button|
		next if button.hidden
		
		#Multi button	
		if button.type =~ /multi/i
			lst_transfo.push [button, xpos, @yref - button.rank * button.buttons[0].height]
			x = xpos
			button.buttons.each do |b|
				next if b.hidden
				bsuite = (b.bsuite) ? b.bsuite : [b]
				bsuite.each do |bs|
					lst_transfo.push [bs, x, @yref]
					x += bs.width
				end	
			end
			xpos += button.width
			cursepa = false
			
		#Other buttons
		else
			rank = button.rank
			if button.sepa
				if cursepa
					button.hidsepa = true
					next
				else
					button.hidsepa = false
					cursepa = true
				end
			else
				cursepa = false
			end			
			if rank != curank
				xrank = 0
			end	
			lst_transfo.push [button, xpos + xrank, @yref - rank * button.height]
			if rank == 0
				xpos += button.width
				curank = nil
				xrank = 0
			else
				xrank += button.width
				curank = rank
			end	
			
		end	
	end
	term = 2
	@total_width = xpos + term
	@total_button_width = @total_width
	@total_height = @height + 3
	
	#Adjusting the tip area if the button zone is too large
	w = vpwid - @total_width
	w = term if w > 0
	@total_width += w
	@button_tip.width = wid_tip + w
	button_make_quads @button_tip
	button_transfo @button_tip, @wid_ctrl + widblason, @yref
	
	#Computing the transformation of all buttons
	lst_transfo.each { |ll| button_transfo ll[0], ll[1] + w, ll[2] }
		
	#Compute the Bandeau and its buttons
	compute_bandeau
	
	#Compute the side palettes
	compute_side_palettes
	
	#Compute the floating palettes
	@lst_floats.each { |floating| compute_floating_palette floating }
	
	#Compute the position of the docking palettes if any
	docking_compute
	
	#Compute the Message band
	@button_message.width = @total_width - @total_width_side
	@button_message.hidden = true if @no_message
	button_make_quads @button_message
	message_transfo @total_width_side
	
	#Compute the background
	compute_quad_background
	compute_quad_background_bandeau
	
	#Computing the list of buttons
	ls_buttons = @hsh_buttons.values.find_all { |button| button_in_family?(button)}	######
	lsb1 = (ls_buttons.find_all { |b| b.selected }).sort { |b1, b2| sort_by_clickable(b1, b2) }
	lsb2 = (ls_buttons.find_all { |b| !b.selected }).sort { |b1, b2| sort_by_clickable(b1, b2) }
	@lst_visible_buttons = lsb2 + lsb1
	
	#Declaring computing OK
	@computed = true
end

#Sort method with clickable button in latest positions
def sort_by_clickable(b1, b2)
	return 1 if b1.selected && !b2.selected
	return -1 if b2.selected && !b1.selected
	return 1 if button_is_clickable?(b1) && !button_is_clickable?(b2)
	return -1 if button_is_clickable?(b2) && !button_is_clickable?(b1)
	-(b1.rank <=> b2.rank)
end

#Set the transformation for the button
def message_transfo(x)
	button = @button_message
	x = @vpwidth - x - button.width unless @side_left
	button.xleft = @tra_basis[12] = x
	button.ybot = @tra_basis[13] = @yref + ((@top_pos) ? button.height : -@height)
	
	tretina = button_retina_transfo(button)
	@tra_basis[5] = -1
	button.tr_tx = tretina * Geom::Transformation.new(@tra_basis)
	@tra_basis[5] = -1
	button.tr = tretina * Geom::Transformation.new(@tra_basis)
	@tra_basis[12] += 1
	@tra_basis[13] -= 1
	button.tr1 = tretina * Geom::Transformation.new(@tra_basis)
	@tra_basis[12] += 1
	@tra_basis[13] -= 1
	button.tr2 = tretina * Geom::Transformation.new(@tra_basis)
end

#Get the Retina transformation for the button
def button_retina_transfo(button)
	fl = button.bfloating
	(fl) ? fl.tretina : @tretina_main
end

#Set the transformation for the button
def button_transfo(button, x, y, noside=false)
	tretina = button_retina_transfo(button)
	x = @vpwidth - x - button.width unless noside || @side_left
	button.xleft = @tra_basis[12] = x
	button.ybot = @tra_basis[13] = y
	@tra_basis[5] = -1
	button.tr_tx = tretina * Geom::Transformation.new(@tra_basis)
	@tra_basis[5] = -1
	button.tr = tretina * Geom::Transformation.new(@tra_basis)
	@tra_basis[12] += 1
	@tra_basis[13] -= 1
	button.tr1 = tretina * Geom::Transformation.new(@tra_basis)
	@tra_basis[12] += 1
	@tra_basis[13] -= 1
	button.tr2 = tretina * Geom::Transformation.new(@tra_basis)
end

def make_quad(w, h, incr=0)
	pt1 = Geom::Point3d.new 0, 0, 0
	pt2 = Geom::Point3d.new 0, h - incr, 0
	pt3 = Geom::Point3d.new w - incr, h - incr, 0
	pt4 = Geom::Point3d.new w - incr, 0, 0
	[pt1, pt2, pt3, pt4]
end

#Draw the background of the palette or bandeau
def compute_quad_background ; compute_quad_background_general(:main) ; end	
def compute_quad_background_bandeau ; compute_quad_background_general(:bandeau) ; end	
def compute_quad_background_general(type)
	if type == :main
		height = @height
		height += @hgt_bandeau
	else
		height = @hgt_bandeau - @dec_bandeau
	end	
	
	if @side_left
		xmin = 0
		xmax = @total_width
	else
		xmax = @view.vpwidth
		xmin = @view.vpwidth - @total_width
	end
	
	if @top_pos
		ymin = 0
		ymax = height + 3
		ylinew = ymax - 1
		ylineb = ylinew + 1
		yminh = ymin
		ymaxh = ylinew
	else
		ymin = @view.vpheight - height - 3
		ymax = @view.vpheight
		ylinew = ymin + 1
		ylineb = ylinew - 1
		yminh = ymax
		ymaxh = ylinew
	end
	quad_background = []
	quad_background.push Geom::Point3d.new(xmin, ymin)
	quad_background.push Geom::Point3d.new(xmax, ymin)
	quad_background.push Geom::Point3d.new(xmax, ymax)
	quad_background.push Geom::Point3d.new(xmin, ymax)
	
	line_background = []
	line_background.push Geom::Point3d.new(xmin, ylinew)
	line_background.push Geom::Point3d.new(xmax, ylinew)
	line_background.push Geom::Point3d.new(xmin, ylineb)
	line_background.push Geom::Point3d.new(xmax, ylineb)
	
	if @side_left
		line_background.push Geom::Point3d.new(xmax-1, yminh)
		line_background.push Geom::Point3d.new(xmax-1, ymaxh)
		line_background.push Geom::Point3d.new(xmax, yminh)
		line_background.push Geom::Point3d.new(xmax, ymaxh)
	else
		line_background.push Geom::Point3d.new(xmin+1, yminh)
		line_background.push Geom::Point3d.new(xmin+1, ymaxh)
		line_background.push Geom::Point3d.new(xmin, yminh)
		line_background.push Geom::Point3d.new(xmin, ymaxh)
	end

	#Retina correction
	line_background = line_background.collect { |pt| @tretina_main * pt }
	quad_background = quad_background.collect { |pt| @tretina_main * pt }
	
	#Assigning the quad and line
	if type == :main
		@line_background = line_background
		@quad_background = quad_background
	else
		@line_background_bandeau = line_background
		@quad_background_bandeau = quad_background
	end	
end

#Compute the Retina transformation for the main palettes
def retina_main_transfo
	xori = (@side_left) ? 0 : @view.vpwidth
	yori = (@top_pos) ? 0 : @view.vpheight
	origin = Geom::Point3d.new xori, yori, 0
	@tretina_main = Geom::Transformation.scaling origin, @retina_factor, @retina_factor, 1	
end

#----------------------------------------------------------------------------
# Drawing methods
#----------------------------------------------------------------------------

#DRAW: Draw the main palette
def draw_main_background(view)
	#draw the background to the main palette
	return if @shrinked
	compute_quad_background unless @quad_background
	view.drawing_color = @bkcolor_normal
	view.draw2d GL_QUADS, @quad_background
	
	view.line_stipple = ''
	view.line_width = 2
	view.drawing_color = @frcolor_dark
	view.draw2d GL_LINE_STRIP, @line_background[0..1]
	view.line_width = 1
	view.drawing_color = @frcolor_white
	view.draw2d GL_LINE_STRIP, @line_background[2..3]

	view.drawing_color = @frcolor_dark
	view.draw2d GL_LINE_STRIP, @line_background[4..5]
	view.drawing_color = @frcolor_white
	view.draw2d GL_LINE_STRIP, @line_background[6..7]
end

#DRAW: Draw the bandeau if applicable
def draw_bandeau_background(view)
	return if @lst_bandeau.empty?
	return if @shrinked
	#return
	
	#Draw the background of the bandeau
	compute_quad_background_bandeau unless @quad_background_bandeau
	view.drawing_color = @bkcolor_bandeau
	view.draw2d GL_QUADS, @quad_background_bandeau
	
	view.line_stipple = ''
	view.line_width = 2
	view.drawing_color = @frcolor_dark
	view.draw2d GL_LINE_STRIP, @line_background_bandeau[0..1]
	view.line_width = 1
	view.drawing_color = @frcolor_white
	view.draw2d GL_LINE_STRIP, @line_background_bandeau[2..3]

	view.drawing_color = @frcolor_dark
	view.draw2d GL_LINE_STRIP, @line_background_bandeau[4..5]
	view.drawing_color = @frcolor_white
	view.draw2d GL_LINE_STRIP, @line_background_bandeau[6..7]
end

#Draw a button in all states
def draw_button(view, button)
	return if button.hidden || button.hidsepa || button.height == 0
	return if (button.bfloating && floating_is_hidden?(button.bfloating)) || (!button.bfloating && @no_main)
	compute_button button
	return unless button.quad && button.tr
	
	#Table button
	return draw_button_table(view, button) if button.table
	
	if button == @button_tip
		return if @button_tip.hidden || (@widmin_tip && @button_tip.width < @widmin_tip)
	end

	bt_color = button.bk_color
	bt_color = bt_color.call if bt_color.class == Proc
	bh_color = button.hi_color
	bh_color = bh_color.call if bh_color.class == Proc
	clickable = button_is_clickable?(button)

	#Drawing the area
	if !clickable
		#bkcolor = @bkcolor_normal
		bkcolor = (bt_color) ? bt_color : @bkcolor_normal
		width = 1
		color1 = @frcolor_dark
		color2 = @frcolor_dark
	elsif button.selected
		bkcolor = (bh_color) ? bh_color : @bkcolor_selected
		width = 2
		color1 = @frcolor_gray
		color2 = @frcolor_white
		if button.frame_hilight
			col = button.frame_hilight
			col = col.call if col.class == Proc
			color1 = color2 = col
		end	
	elsif button == @sel_button	
		bkcolor = @bkcolor_down
		width = 2
		color1 = @frcolor_gray
		color2 = @frcolor_white
	elsif button == @hi_button && !(button.no_hilight || button.frame_hilight)
		bkcolor = (bt_color) ? bt_color : @bkcolor_hilight
		width = 2
		color1 = @frcolor_white
		color2 = @frcolor_gray
	else	
		bkcolor = (bt_color) ? bt_color : @bkcolor_normal
		width = 1
		color1 = 'darkgray'
		color2 = 'darkgray'
	end	
	bkcolor = 'lightgrey' if button.grayed
	
	#Framing for the button
	pts11 = pts21 = nil
	if width == 1
		t = button.tr
		quad = button.quad
		pts = quad.collect { |pt| t * pt }
		pts1 = [pts[1], pts[2]]
		pts2 = [pts[0], pts[3]]
	elsif button.selected && button.frame_hilight
		t = button.tr2
		quad = button.quad2
		pts = quad.collect { |pt| t * pt }
		pts1 = [pts[0], pts[1], pts[2]]
		pts2 = [pts[2], pts[3], pts[0]]
		t = button.tr1
		quad = button.quad1
		pts = quad.collect { |pt| t * pt }
		pts11 = [pts[0], pts[1], pts[2]]
		pts21 = [pts[2], pts[3], pts[0]]
	else
		t = button.tr1
		quad = button.quad1
		pts = quad.collect { |pt| t * pt }
		pts1 = [pts[0], pts[1], pts[2]]
		pts2 = [pts[2], pts[3], pts[0]]
	end	

	#Drawing the background
	unless button.tab_style
		view.drawing_color = bkcolor
		view.draw2d GL_QUADS, pts
		content = draw_button_content view, button
		bdinfo = button.border_color
		bdinfo = bdinfo.call if bdinfo.class == Proc
		if bdinfo
			bdinfo = [bdinfo] unless bdinfo.class == Array
			col, wid, stipple = bdinfo
			view.drawing_color = col if col
			view.line_width = (wid) ? wid : 1
			view.line_stipple = (stipple) ? stipple : ''
			t = button.tr1
			pts_bd = button.quad1.collect { |pt| t * pt }
			view.draw2d GL_LINE_LOOP, pts_bd			
		end
	end
	
	#Drawing the border
	if button.tab_style
		draw_button_tab view, button, bt_color
	elsif width > 1
		view.line_width = width
		view.line_stipple = ""
		view.drawing_color = color1
		view.draw2d GL_LINE_STRIP, pts1
		view.drawing_color = color2
		view.draw2d GL_LINE_STRIP, pts2
	end
				
	#Compute the text
	text = button.text
	if button.text_proc
		text = button.text_proc.call
	end
	
	#Draw a border for groups
	if clickable && width == 1 && !content && button.text == nil
		view.drawing_color = 'silver'
		view.line_width = 1
		pts = button.quad2.collect { |pt| t * pt }
	end	
	
	underline = (!clickable || (!button.selected && button != @hi_button)) && !button.bk_color && button.rank > 0 && 
	   (button.text || button.text_proc) && button.draw_proc == nil && button.draw_code == nil
	   
	t = button.tr1 if underline && !button.tab_style
	
	#Draw the text if any
	if text
		draw_button_text view, button, text
	end
	
	#Draw a separator for title
	if underline
		view.line_width = 1
		view.line_stipple = ""
		pts1 = [button.quad2[0], button.quad2[3]].collect { |pt| t * Geom::Point3d.new(pt.x, pt.y+1, 0) }
		pts2 = [button.quad2[0], button.quad2[3]].collect { |pt| t * pt }
		view.drawing_color = @bkcolor_title1_u
		view.draw2d GL_LINE_STRIP, pts1
		view.drawing_color = @bkcolor_title1_d
		view.draw2d GL_LINE_STRIP, pts2
	end		
end

#Draw a table button
def draw_button_table(view, button)
	button.table.draw view, button.tr
end

#Draw Button Text
def draw_button_text(view, button, text)
	#Which text to draw	 
	text = (button.selected && text[1]) ? text[1] : text[0] if text.class == Array
	return unless text

	#Retrieving the text information
	if button.selected
		instructions = button.text_sel_instructions
		text_previous = button.text_sel_previous
	else
		instructions = button.text_instructions
		text_previous = button.text_previous
	end
	
	#Computing the instructions
	unless instructions && text == text_previous
		wid = button.width
		hgt = button.height
		y = button.height+1
		y -= 1 if button.tab_style
		x = 3
		justif = (button.justif) ? button.justif : 'MT' 
		instructions = G6.text_box_instructions(text, x, y, wid, hgt, justif)
	end
		
	#Storing the text information	
	if button.selected
		button.text_sel_instructions = instructions
		button.text_sel_previous = text_previous
	else
		button.text_instructions = instructions
		button.text_previous = text_previous
	end

	#Drawing the text
	t = button.tr_tx
	instructions.each do |lst|
		s, pt = lst
		G6.view_draw_text view, t * pt, s
	end	
end

#Draw buttons with TAB style
def draw_button_tab(view, button, bt_color)
	bkcolor = (bt_color) ? bt_color : @bkcolor_normal
	bkcolor = 'silver' if button == @hi_button
	bkcolor = @bkcolor_down if button.selected
	
	t = (button.selected) ? button.tr : button.tr
	quad = (button.selected) ? button.quad1 : button.quad2
	dec = 3
	pts = quad.collect { |pt| t * pt }
	pts[2].x -= 1
	pts[3].x -= 1
	ptc10 = Geom::Point3d.new pts[1].x, pts[1].y+dec, 0
	ptc12 = Geom::Point3d.new pts[1].x+dec, pts[1].y, 0
	ptc21 = Geom::Point3d.new pts[2].x-dec, pts[2].y, 0
	ptc23 = Geom::Point3d.new pts[3].x, pts[2].y+dec, 0
	lptw = [pts[0], ptc10, ptc12, ptc21]
	lptw = [pts[3], pts[0]] + lptw unless button.selected
	lptg = [ptc21, ptc23, pts[3]]
	
	poly = [pts[0], ptc10, ptc12, ptc21, ptc23, pts[3], pts[0]]
	view.drawing_color = bkcolor
	view.draw2d GL_POLYGON, poly
	
	draw_button_content view, button
	
	view.line_stipple = ""
	view.line_width = 1
	view.drawing_color = @frcolor_gray
	view.draw2d GL_LINE_STRIP, lptg
	if button.selected
		view.line_width = 2
		view.draw2d GL_LINE_STRIP, [ptc23, pts[3]]
	end
	view.line_width = 1
	view.drawing_color = @frcolor_white
	view.draw2d GL_LINE_STRIP, lptw
end

def draw_button_content(view, button)
	#button colors
	if button.grayed
		main_color = @frcolor_white
		frame_color = @frcolor_dark
	else
		main_color = button.main_color
		frame_color = button.frame_color
	end
		
	#Getting the instructions
	if button.grayed
		instructions = button.gray_instructions
		unless instructions && !button.draw_refresh	
			instructions = []
			if button.draw_code
				codes = button.draw_code
				codes = [codes] if codes.class != Array
				codes.each do |code|
					instructions += @standard_draw_proc.call(code, button.width - 4, 
															 button.height - 4, main_color, frame_color,
															 button.draw_scale, button.draw_extra_proc, button.selected, true)
				end													
			end
			instructions += compute_instructions(button, main_color, frame_color)
			instructions += fkey_instructions(button) if button.fkey
			button.gray_instructions = instructions
		end

	elsif button.selected
		instructions = button.sel_instructions
		unless instructions && !button.draw_refresh	
			instructions = []
			if button.draw_code
				codes = button.draw_code
				codes = [codes] if codes.class != Array
				codes.each do |code|
					instructions += @standard_draw_proc.call(code, button.width - 4, 
															 button.height - 4, main_color, frame_color,
															 button.draw_scale, button.draw_extra_proc, true)
				end													
			end
			instructions += compute_instructions(button, main_color, frame_color)
			instructions += fkey_instructions(button) if button.fkey
			button.sel_instructions = instructions
		end
	
	else
		instructions = button.instructions
		unless instructions && !button.draw_refresh	
			instructions = []
			instructions += blason_instructions(button, main_color, frame_color) if button.blason
			if button.draw_code
				codes = button.draw_code
				codes = [codes] if codes.class != Array
				codes.each do |code|
					instructions += @standard_draw_proc.call(code, button.width - 4, 
															 button.height - 4, main_color, frame_color,
															 button.draw_scale, button.draw_extra_proc, false)
				end													
			end
			instructions += compute_instructions(button, main_color, frame_color)
			instructions += fkey_instructions(button) if button.fkey
			button.instructions = instructions
		end
	end
	
	return false unless instructions
	
	#Processing the instructions
	@ogl.process_draw_GL view, button.tr2, instructions
end

#Compute the instructions for the button
def compute_instructions(button, main_color, frame_color)
	draw_proc = button.draw_proc
	return [] unless draw_proc
	
	case draw_proc.arity
	when 3
		draw_proc.call(button.symb, button.width - 4, button.height - 4)
	when 4
		draw_proc.call(button.symb, button.width - 4, button.height - 4, button.selected)
	when 5
		draw_proc.call(button.symb, button.width - 4, button.height - 4, main_color, frame_color)
	when 6
		draw_proc.call(button.symb, button.width - 4, button.height - 4, main_color, frame_color, button.selected)
	when 7
		draw_proc.call(button.symb, button.width - 4, button.height - 4, main_color, frame_color, button.selected, button.grayed)
	end	
end

#Draw the frame for a blason
def blason_instructions(button, main_color, frame_color)
	main_color = (button.main_color) ? button.main_color : @bkcolor_blason unless main_color
	frame_color = (button.frame_color) ? button.frame_color : @frcolor_blason unless frame_color
	@standard_draw_proc.call :std_blason, button.width - 4, button.height - 4, main_color, frame_color, false
end

#Draw the frame for a blason
def fkey_instructions(button)
	main_color = 'green'
	x = button.width - 4
	y = 1
	frame_color = (button.frame_color) ? button.frame_color : @frcolor_blason unless frame_color
	@ogl.digit_instructions button.fkey.fkey, x, y, main_color
end

#Top drawing method for the palette
def draw(view)
	return if @palette_whole_hidden
	
	#List of all visible buttons
	compute_all unless @computed
	ls_buttons = list_visible
		
	#Tracking all changes since last refresh
	ls_buttons.each { |button| button_track_changes button }
	
	#Recomputing the palette if necessary
	compute_all unless @once
		
	#Drawing the background for the Main palettes
	unless @no_main
		draw_main_background view
		draw_bandeau_background view
	end	
	
	#Drawing the background for the floating palettes
	if !@shrinked || @no_main
		@lst_floats.each do |floating|
			next if floating == @dropdown_cur
			draw_floating_background view, floating
		end
	end
	
	#Drawing the elements
	ls_last = []
	lst_dropbut = []
	ls_buttons.each do |button| 
		if button.zone_proc
			ls_last.push button
		elsif @dropdown_cur && button.bfloating == @dropdown_cur
			lst_dropbut.push button
		else	
			draw_button view, button
		end	
	end	
	
	#Highlight button
	hi_button = hi_dp_button = nil
	if @hi_button && button_is_clickable?(@hi_button)
		if @dropdown_cur && @hi_button.bfloating == @dropdown_cur
			hi_dp_button = @hi_button
		else
			hi_button = @hi_button
		end	
	end
	
	#Drawing the latest buttons
	ls_last.each { |button| draw_button view, button }
	
	#Drawing the highlight button
	draw_button view, hi_button if hi_button
	
	#Drawing the drop-down palette if any
	if @dropdown_cur
		draw_floating_background view, @dropdown_cur
		lst_dropbut.each { |button| draw_button view, button }
		draw_button view, hi_dp_button if hi_dp_button
	end
	
	#Post drawing
	post_draw view
end

#Drawing after the palette
def post_draw(view)
	@post_drawing_buttons.each do |button|
		next if button.hidden
		button.post_draw_proc.call @view, button.symb, @xmove, @ymove
	end	
end

#----------------------------------------------------------------------------
# Change of state
#----------------------------------------------------------------------------

def button_track_changes(button)
	if button.grayed != button_check_grayed?(button)
		button.instructions = nil
	end	
	if button.hidden != button_check_hidden?(button)
		@computed = false
	end	

end

#Test the status Grayed of a button
def button_check_grayed?(button)
	if button.parent
		button.grayed = button_check_grayed?(button.parent)
	end
	button.grayed = button.grayed_proc.call if button.grayed_proc
	button.grayed
end

#Test the status Grayed of a button
def button_check_hidden?(button)
	if button.parent
		button.hidden = button_check_hidden?(button.parent)
	end
	button.hidden = button.hidden_proc.call if button.hidden_proc
	button.hidden
end
	
#Check if a button is hidden	
def button_hidden?(button_symb)
	button = @hsh_buttons[button_symb]
	(button) ? button_check_hidden?(button) : true
end

#----------------------------------------------------------------------------
# Message Zone
#----------------------------------------------------------------------------

#Set the tooltip
def set_tooltip(tooltip=nil, level=nil)
	text = (tooltip) ? tooltip : ""
	set_message_and_tip @button_tip, text, level
end

#Set a Message
def set_message(message=nil, level=nil)
	@shadow_bug_displayed = false
	text = (message) ? message.gsub(/\n/, ' - ') : nil
	if @no_message
		Sketchup.set_status_text((text) ? text : "")
	else
		set_message_and_tip @button_message, message, level
		@button_message.hidden = (text) ? false : true
	end	
end

#Utility to set Message and Tip
def set_message_and_tip(button, text=nil, level=nil)
	button.text = text
	case level
	when 'W', 'w'
		color = 'yellow'
	when 'H', 'h'
		color = 'lightyellow'
	when 'I', 'i'
		color = 'lightgreen'
	when 'E', 'e'
		color = 'lightcoral'
	when /..+/
		color = level
	else
		color = nil
	end
	button.bk_color = color
end

#----------------------------------------------------------------------------
# Palette placement
#----------------------------------------------------------------------------

#Turn off the palette
def turn_on ; @palette_whole_hidden = false ; end
def turn_off ; @palette_whole_hidden = true ; @mouse_button = nil ; end

#Toggle shrink or expand of palettes
def toggle_shrinked
	@shrinked = !@shrinked
	refresh_after_toggle true
end

#Toggle position top or bottom of the palettes
def toggle_top_down
	@top_pos = !@top_pos
	notify_change_position
	refresh_after_toggle
end

#Toggle visibility of the message zone
def toggle_message_visible
	@message_visi = !@message_visi
	refresh_after_toggle
end

#Toggle position left or right of the palettes
def toggle_side_left
	@side_left = !@side_left
	notify_change_position
	refresh_after_toggle
end

#Notification of a change of position of the palette
def notify_change_position
	@notify_change_position_proc.call if @notify_change_position_proc
end

#Refresh after changing the palette position
def refresh_after_toggle(recompute=true)
	retina_main_transfo
	adjust_draw_standard_button
	persistence_save
	if recompute
		@computed = false
		compute_all
	end	
	onMouseMove_zero
	@view.invalidate
end

#----------------------------------------------------------------------------
# Mouse handling
#----------------------------------------------------------------------------

#check if a button is not clickable
def button_is_clickable?(button)
	!(button_passive?(button) || button.grayed || button.hidden || button.hidsepa || button.table)
end

def button_passive?(button)
	passive = button.passive
	(passive.class == Proc) ? passive.call : passive
end

#Return the list of visible buttons
def list_visible
	if (@shrinked)
		ls = [@button_pos, @button_shrink, @button_msg_visi, @button_sideleft].find_all { |b| !b.hidden }
	else
		compute_all unless @lst_visible_buttons
		ls = @lst_visible_buttons
	end
	
	#Put the buttons of the drop down palette in first position
	if @dropdown_cur
		ls1 = (ls.find_all { |b| b.bfloating == @dropdown_cur })
		ls2 = (ls.find_all { |b| b.bfloating != @dropdown_cur })
		ls = ls1 + ls2
	end
	
	ls
end

#Check if a given position (x, y) is within a given button
def mouse_within_button?(button_symb, x, y)
	button = @hsh_buttons[button_symb]
	(@mouse_button == button)
end

#Determine if an active button is under the mouse
def locate_button(x, y)
	#Quick test if within the palette
	@mouse_button = nil
	@within_background = nil
	ptxy = Geom::Point3d.new x, y, 0
	
	#Finding the button
	list_visible.each do |button|
		next unless button_visible?(button) && button.xleft && button_in_family?(button)
		pt = button_retina_transfo(button).inverse * ptxy
		xm = pt.x
		ym = pt.y
		if (xm >= button.xleft && xm <= button.xleft + button.width) && 
		   (ym <= button.ybot + 2 && ym >= button.ybot - button.height - 3 )
			@mouse_button = button
			table_locate_button(button, ptxy) if button.table
			break
		end	
	end
	@mouse_button
end

#check if a button is visible
def button_visible?(button)
	return false if button.hidden || button.sepa
	return false if button.blason && @no_main
	bfloating = button.bfloating
	return false if bfloating && floating_is_hidden?(bfloating)
	true
end

#Detrmine if the mouse if still within the bounday of the main palette
def locate_within_background?(ptxy)
	return false if @no_main
	@within_background = Geom.point_in_polygon_2D(ptxy, @quad_background, true)
end

#Set the cursor
def onSetCursor
	return false if @palette_whole_hidden
	return false unless @mouse_button || @mouse_floating || @within_background
	if @mouse_floating
		ic = (@mouse_floating_frame) ? @id_cursor_move : @id_cursor_move
	elsif @button_zone && @button_zone.cursor_proc
		ic = @button_zone.cursor_proc.call 
		ic = @id_cursor_arrow unless ic
	elsif @mouse_button && button_is_clickable?(@mouse_button)
		ic = @id_cursor_hand
	else
		ic = @id_cursor_arrow
	end
	UI::set_cursor ic
	ic
end

#Return the current mouse position
def current_mouse_position
	[@xmove, @ymove]
end

#Mouse move methods
def onMouseMove_zero
	onMouseMove -1, @ptxy.x, @ptxy.y, @view
end

def onMouseMove(flags, x, y, view)
	@view.tooltip = @main_tooltip = ""
	return false if @palette_whole_hidden
	return false if cancel_button_down(flags) || !@ptxy || !x || !y
	@xmove = @ptxy.x = x
	@ymove = @ptxy.y = y
	
	#Handling Button highlight
	button = locate_button x, y
	
	#Button_zone out
	if @button_zone && button != @button_zone
		manage_zone_button(:out, @button_zone, x, y)
		if @button_down
			manage_zone_button(:move, @button_zone, x, y)
		else
			manage_zone_button(:out, @button_zone, x, y)
			@button_zone = nil
		end
		return true		
	end	
		
	#Handling moving of floating palettes
	if test_auto_hide_floating(x, y)
		@view.invalidate
		return false
	end
	
	if @origin_floating
		move_current_floating x, y
		return true
	end	
	
	#Follow mouse floating palette
	follow_mouse_floating x, y
	
	#Button Zone
	#manage_zone_button(:move, button, x, y)
	
	#Check for floating palette
	unless button || @button_down
		@hi_button = nil
		return true if locate_floating_frame(@ptxy)
		if locate_within_background?(@ptxy)
			###view.invalidate
			return true 
		end	
		return false
	end	
	@mouse_floating = nil
	
	#Button is actually down but flag is not on
	return true if flags & 1 == 1 && !@button_down
		
	@hi_button = nil
	
	#Special case for dropdown
	if @button_abs_down	&& button && button.bfloating && button.bfloating == @dropdown_cur
		@hi_button = button
	elsif @button_down
		#otherwise, ignore	
		if button == nil || (@sel_button && button != @sel_button)
			return true
		end
	else	
		@hi_button = button
	end	
	
	#Button Zone
	manage_zone_button(:move, button, x, y)
	
	#Compute the tooltip is button is highlighted
	if @hi_button
		tip = button_compute_tooltip(@hi_button)
		@main_tooltip = (tip) ? tip.gsub(/\n/, ' - ') : ""
		@view.tooltip = (tip) ? tip : ""
		@hi_button.hover_proc.call if @hi_button.hover_proc
		###@view.invalidate
	end	
	
	true
end

#Notify a zone button about mouse events
def manage_zone_button(code, button, x, y)
	return false unless button && button.zone_proc && !button_check_grayed?(button)
	@button_zone = button
	ptxy = button_retina_transfo(button).inverse * Geom::Point3d.new(x, y, 0)
	button.zone_proc.call code, @button_down, button.width-4, button.height-4, ptxy.x - button.xleft, button.ybot - ptxy.y
	true
end

#Mouse enters the view port
def onMouseLeave(view)
	@hi_button = false
	@view.invalidate
end

#Mouse leaves the view port
def onMouseEnter(view)
end

def button_down?
	@button_down
end
	
#Cancel a button down event when Up event is not in the same area
def cancel_button_down(flags)
	return false
	if (flags == nil || flags >= 0) && @button_down && (flags & 1 != 1)
		@hi_button = nil
		@sel_button = nil
		@button_down = false
		@view.invalidate
	end	
	false
end	

#Button click - Means that we end the selection
def onLButtonDown(flags, x, y, view)
	@button_abs_down = true
	return false if @palette_whole_hidden
	@xdown = x
	@ydown = y
	@ptxy.x = x ; @ptxy.y = y
	@time_down = nil
	
	#Handling the floating palette
	if @origin_floating
		@origin_floating = nil
		@view.invalidate
		end_move_floating
		return true
	end
	
	if @mouse_floating
		@origin_floating = [@mouse_floating.xpos - x, @mouse_floating.ypos - y]
		@view.invalidate
		return true
	end
	
	#Handling button click
	button = locate_button x, y
	unless button
		dropdown_finish
		return true if locate_within_background?(@ptxy)
		return false
	end	
	@button_down = true
	@view.tooltip = @main_tooltip = ""
	@sel_button = button
	@sel_button = nil unless button_is_clickable?(@sel_button)
	@time_down = Time.now.to_f if button
	
	#Button with drop down attached
	dropdown_process_down_from_button(button)
	
	#Button with zone attached or button not clickable
	return true if manage_zone_button(:button_down, button, x, y) || !@sel_button
		
	#Button with drop down attached
	#dropdown_process_down_from_button(button)
	
	#Floating palette with drop_start procedure
	
	###@view.invalidate
	true
end

#Button click - Means that we end the selection
def onLButtonUp(flags, x, y, view)
	@button_abs_down = false
	return false if @palette_whole_hidden

	#Handling the floating palette
	@long_click = nil
	@ptxy.x = x ; @ptxy.y = y
	if @origin_floating 
		@origin_floating = nil if ((@xdown - x).abs > 4 || (@ydown - y).abs > 4)
		@view.invalidate
		end_move_floating unless @origin_floating
		return true
	elsif @mouse_floating
		return true
	end

	#Handling button click
	button = locate_button x, y
	
	#Button Zone on and no button selected
	if @button_zone && !button
		manage_zone_button(:out, @button_zone, x, y)
		@button_zone = nil
	end
	
	#No button selected - finish drop-down mode if any active
	dropdown_finish unless button || (@dropdown_cur && @dropdown_cur.invoked)
	
	#Mismatch of up and down
	if button && !@button_down
		@hi_button = nil
		@view.invalidate
		return true
	end
	
	#Not an event for the palette
	unless @button_down && button
		@button_down = false
		@hi_button = nil
		@sel_button = nil
		@view.invalidate
		return true if locate_within_background?(@ptxy)
		automove_floating x, y
		return false
	end
	
	@button_down = false
	
	#Checking for dragging on drop-down
	if dropdown_match_up(button) && ((@xdown - x).abs > 6 || (@ydown - y).abs > 6)
		execute_button button if button_is_clickable?(button)
	
	#Trapping the click if click up is on the same button as click down
	else (@sel_button && button == @sel_button) || button == @button_zone
		@long_click = Time.now.to_f - @time_down if @time_down
		if button == @button_zone
			code = (long_click? && ((@xdown - x).abs < 4 && (@ydown - y).abs < 4)) ? :long_click : :button_up
			manage_zone_button(code, button, x, y)
		elsif @sel_button
			execute_button @sel_button if button_is_clickable?(@sel_button)
		end	
	end

	#dropdown_finish
	@sel_button = nil
	@view.invalidate
	true
end

def onLButtonDoubleClick(flags, x, y, view)
	return false if @palette_whole_hidden
	@ptxy.x = x ; @ptxy.y = y
	button = locate_button x, y
	status = false
	if button
		if button == @button_zone
			manage_zone_button(:double_click, button, x, y)
			status = true
		elsif button.double_click_proc
			button.double_click_proc.call
			status = true
		else
			status = onLButtonDown(flags, x, y, view)
			onLButtonUp(flags, x, y, view)
		end	
	else
		status = true if locate_within_background?(@ptxy)
	end	
	@view.invalidate
	status
end

#Test if the action was generated from a long click
def long_click?(duration=nil)
	duration = @long_click_duration unless duration
	(@long_click && @long_click > duration)
end

#----------------------------------------------------------------------------
# DROPDOWN: Dropdown Management
#----------------------------------------------------------------------------

#DROPDOWN: Notify the owner of the dropdown
def dropdown_notify(flpal, *args)
	flpal.dropdown_proc.call *args if flpal.dropdown_proc.class == Method
end

#DROPDOWN: Expand the drop-down when Click is down
def dropdown_invoke(flpal_symb)
	#Retrieving the floating palette
	flpal = @hsh_floats[flpal_symb]
	return unless flpal
	
	#Simulating button down
	@button_down = @button_abs_down
	
	#Setting the current drop-down with the floating palette
	dropdown_finish
	@dropdown_cur = flpal
	flpal.invoked = true
	
	#Notifying the floating palette
	dropdown_notify flpal, :show
	
	#Setting the position of the drop-down
	@computed = false
	dropdown_set_position	
end

#DROPDOWN: Expand the drop-down when Click is down
def dropdown_process_down_from_button(button_from)
	#Terminate drop-down for other buttons
	if @dropdown_cur && button_from && button_from.dropdown_pal == @dropdown_cur.symb
		dropdown_finish
		return true
	end	
	if @dropdown_cur && button_from && button_from.bfloating != @dropdown_cur
		dropdown_finish
	end	

	#Checking if button has a drop-down attached
	flpal_symb = button_from.dropdown_pal
	return false unless flpal_symb
	flpal = @hsh_floats[flpal_symb]
	return false unless flpal
	
	#Setting the current drop-down with the floating palette
	dropdown_finish
	@dropdown_cur = flpal
	
	#Notifying the floating palette
	dropdown_notify flpal, :show, ((button_from) ? button_from.symb : nil)
	
	#Setting the position of the drop-down
	@computed = false
	dropdown_set_position(button_from)
	true	
end

#DROPDOWN: Compute the position of the drop-down palette
def dropdown_set_position(button_from=nil)
	#Position determined by the mouse
	unless button_from
		floating_set_position(@dropdown_cur.symb, @xmove, @ymove)	
		return
	end
		
	#Get the standard position
	tr = button_from.tr
	quad = button_from.quad
	
	iquad = 0
	pt1 = tr * quad[iquad]
	xpos = pt1.x
	if @top_pos
		ypos = pt1.y
	else
		@dropdown_cur.buttons.each { |button| compute_button button, true }
		compute_floating_palette @dropdown_cur
		quad_out = @dropdown_cur.quad_out
		ptf1 = quad_out[0]
		ptf2 = quad_out[3]
		h = (ptf1.y - ptf2.y).abs
		pt2 = tr * quad[1]
		ypos = pt2.y - h
	end

	#Setting the position
	floating_set_position(@dropdown_cur.symb, xpos, ypos)
end

#DROPDOWN: Check if there is an active drop-down
def dropdown_active?
	(@dropdown_cur) ? true : false
end
	
#DROPDOWN: Close the active drop-down
def dropdown_finish
	return unless @dropdown_cur
	#ddproc = @dropdown_cur.dropdown_proc
	#ddproc.call :hide, nil if ddproc.class == Method
	dropdown_notify @dropdown_cur, :hide
	@dropdown_cur.invoked = true
	@dropdown_cur = nil
end
	
#DROPDOWN: Check if UP event is applicable to the drop-down palette
def dropdown_match_up(button)
	return false unless @dropdown_cur && button
	if @sel_button
		dropdown_pal = @sel_button.dropdown_pal
		return false unless dropdown_pal == @dropdown_cur.symb && button.floating == dropdown_pal
	else	
		return false unless button.bfloating == @dropdown_cur
	end	
	true
end
	
#----------------------------------------------------------------------------
# EXEC: Execution of actions
#----------------------------------------------------------------------------

#EXEC: Execution of action when clicking on button
def execute_button(button)
	return unless button && button_is_clickable?(button)
	
	symb = button.symb
	
	#Standard
	if symb == :_bt_pos
		return toggle_top_down
	elsif symb == :_bt_shrink
		return toggle_shrinked
	elsif symb == :_bt_msg_visi
		return toggle_message_visible
	elsif symb == :_bt_sideleft
		return toggle_side_left
	end
	
	#Checking multiple button
	parent = button.parent
	if parent && parent.multi
		execute_multi button, parent
		button = parent
	elsif button.multi
		button.value = button.default_value	
	end
	
	#Calling the custom proc
	if button.long_click_proc && long_click?
		button.long_click_proc.call
	elsif button.action_proc
		button.action_proc.call
	elsif @palette_proc	
		@palette_proc.call :click, button.symb, button.value
	end	
end

#EXEC: Execution method parent button controlling multi-choice buttons
def execute_multi(button, parent)
	sval = parent.value
	if parent.radio
		parent.value = button.value
	else
		lv = (sval) ? sval.split(';;') : []
		if button.selected
			lv -= [button.value]
		else
			lv.push button.value unless lv.include?(button.value)
		end
		parent.value = lv.join ';;'
	end	
end
	
#----------------------------------------------------------------------------
# TABLE: Table button management
#----------------------------------------------------------------------------

#TABLE: Declare a table button
def table_create(*hspecs)
	table = PaletteTable.new self, *hspecs
end	
	
#TABLE: Compute the layout of the button table	
def table_compute_layout(table_button)
	table = table_button.table
	
	#Sub buttons
	wtot = 200
	htot = 100
	info = []
	table_button.buttons.each do |button|
		w = button.width
		h = button.height
		puts "sub button = #{button.symb} w = #{button.width} h = #{button.height}"
		wtot = w if w > wtot
		htot = h if h > htot
		info.push [button, 0, 0]
	end
	table_button.buttons_info_pos = info
	
	#Returning the size of the table button
	puts "Compute table button"
	table_button.width = wtot
	table_button.height = htot
		
end
	
#TABLE: Final positioning of the buttons within a table button	
def table_position_buttons(table_button, decx, decy)
	hgt = table_button.height
	table_button.buttons_info_pos.each do |button, x, y|
		h = button.height
		yt = y + decy - hgt + h
		button_transfo button, x + decx, yt, true
	end	
end

#TABLE: Locate a sub button within a table button	
def table_locate_button(table_button, ptxy)
	table_button.buttons_info_pos.each  do |button, x, y|
		pt = button_retina_transfo(button).inverse * ptxy
		xm = pt.x
		ym = pt.y
		if (xm >= button.xleft && xm <= button.xleft + button.width) && 
		   (ym <= button.ybot + 2 && ym >= button.ybot - button.height - 3 )
			@mouse_button = button
			return button
		end			
	end
	nil
end
	
end	# class Palette

#*************************************************************************
#*************************************************************************
# SLIDER: Palette slider button for animation
#*************************************************************************
#*************************************************************************

class PaletteSliderButton

#SLIDER: Class instance initialization and register button in palette
def initialize(palette, button_symb, hslider_specs, *hargs)
	#Initialization
	@palette = palette
	@button_symb = button_symb
	
	#Parameters of the sliders
	@cursor = Traductor.create_cursor "Cursor_Hand", 6, 0	
	@pix_precision = 8
	@color_cursor = 'lightgrey'
	@color_cursor_exact = 'lightgreen'
	@color_cursor_def = 'lightblue'
	@bk_color_text = 'lightgreen'
	@fr_color_text = 'gray'
	@fr_color_text_exact = 'green'

	#Parsing the specs
	hslider_specs.each do |key, value|
		case key
		when :notify_proc
			@notify_proc = value
		when :input
			@input = value
		when :style
			@style = value
		when :cursor
			@cursor = value
		when :pix_precision
			@pix_precision = value
		end
	end	
	
	#Parameters based on style
	case @style
	when :movie
		@progression_dec = 8
	else
		@progression_dec = 3
	end
	
	#Notification proc
	if !@notify_proc && @input
		@notify_proc = self.method("input_slider_proc")
	end
	
	#Button registration
	draw_progression_proc = self.method "instructions"
	zone_progression_proc = self.method "zone_event"
	post_draw_proc = self.method "draw_time"
	
	hsh_spec = {}
	hargs.each { |harg| hsh_spec.update harg if harg.class == Hash }	
	
	cursor_proc = proc { @cursor }
	hsh = { :passive => true, :draw_proc => draw_progression_proc, :draw_refresh => true,
	        :zone_proc => zone_progression_proc, :cursor_proc => cursor_proc, :post_draw_proc => post_draw_proc }
			
	@palette.declare_button(button_symb, hsh, hsh_spec)
end

#SLIDER: Dynamic calculation of the graphics for the progression bar
def instructions(symb, dx, dy, main_color, frame_color, selected, grayed)
	case @style
	when :movie
		instructions_movie(symb, dx, dy, main_color, frame_color, selected, grayed)
	else
		instructions_ruler(symb, dx, dy, main_color, frame_color, selected, grayed)
	end	
end

#SLIDER: RULER style
def instructions_ruler(symb, dx, dy, main_color, frame_color, selected, grayed)
	in_zone = (@code_progression && @code_progression != :out)
	
	#Initialization
	lst_gl = []
	dx2 = dx/2
	dy2 = dy/2
	dec = @progression_dec
	hgt = dy * 0.5
	wid = hgt * 0.6
	ybase = -1
	htick = 4
	htick_def = 6
	xmin = dec
	xmax = dx - dec
	@xlen = xmax - xmin
	
	#Colors
	fr_color_w = 'whitesmoke'
	fr_color_g = 'gray'
	color_tick = (in_zone) ? 'black' : 'gray'
	color_tick_def = (in_zone) ? 'blue' : 'gray'
	
	#Main lines
	yline = htick + hgt / 2 + 2
	line = [Geom::Point3d.new(xmin, yline+1), Geom::Point3d.new(xmax, yline+1)]
	lst_gl.push [GL_LINE_STRIP, line, fr_color_g, 2, '']
	line = [Geom::Point3d.new(xmin, yline-1), Geom::Point3d.new(xmax, yline-1)]
	lst_gl.push [GL_LINE_STRIP, line, fr_color_w, 2, '']
	
	#Getting the current value and range
	vmin = @notify_proc.call(:value_min)
	vmax = @notify_proc.call(:value_max)
	range = vmax - vmin
	current_value = @notify_proc.call(:current_value)
	return lst_gl unless current_value && range && range > vmin	
	
	#Drawing the ticks
	vincr = @notify_proc.call(:value_ticks)
	vdef = @notify_proc.call(:value_default)
	lines = []
	lines_def = []
	if vincr && vincr.class == Array
		vincr.each do |v|
			ratio = (v - vmin) / range
			next if ratio > 1 || ratio < 0
			x = xmin + @xlen * ratio
			if v == vdef
				lines_def.push Geom::Point3d.new(x, ybase), Geom::Point3d.new(x, ybase + htick_def)
			else	
				lines.push Geom::Point3d.new(x, ybase), Geom::Point3d.new(x, ybase + htick)
			end
		end
	elsif vincr.is_a?(Numeric)
		v = vmin
		while true
			ratio = (v - vmin) / range
			next if ratio < 0
			break if ratio > 1
			x = xmin + @xlen * ratio
			line = [Geom::Point3d.new(x, ybase), Geom::Point3d.new(x, ybase + htick)]
			(v == vdef) ? lines_def.concat(line) : lines.concat(line)
			v += vincr
		end		
	end
	lst_gl.push [GL_LINES, lines, color_tick, 1, ''] unless lines.empty?
	lst_gl.push [GL_LINES, lines_def, color_tick_def, 2, ''] unless lines_def.empty?
	
	#Drawing the value cursor
	ratio = (current_value - vmin) / range	
	vround = value_rounding(ratio)[1]
	if vround
		color_cursor = (vround == vdef) ? @color_cursor_def : @color_cursor_exact
	else
		color_cursor = @color_cursor
	end	
	xpos = xmin + @xlen * ratio
	ytop = yline + hgt - 1
	ybot = yline - hgt
	xbeg = xpos - wid
	xend = xpos + wid
	pt1 = Geom::Point3d.new xend, ytop
	pt2 = Geom::Point3d.new xbeg, ytop
	pt3 = Geom::Point3d.new xbeg, ytop - hgt * 1.2
	pt4 = Geom::Point3d.new xpos, ybot
	pt5 = Geom::Point3d.new xend, pt3.y
	pt6 = Geom::Point3d.new xend, ytop
	lst_gl.push [GL_POLYGON, [pt1, pt2, pt3, pt4, pt5, pt6], color_cursor]
	lst_gl.push [GL_LINE_STRIP, [pt1, pt2, pt3, pt4], fr_color_w, 2, '']
	lst_gl.push [GL_LINE_STRIP, [pt4, pt5, pt6], fr_color_g, 2, '']
		
	lst_gl
end

#SLIDER: Movie style
def instructions_movie(symb, dx, dy, main_color, frame_color, selected, grayed)
	in_zone = (@code_progression && @code_progression != :out)
	
	#Initialization
	lst_gl = []
	dx2 = dx/2
	dy2 = dy/2
	dec = @progression_dec
	ywid = (in_zone) ? 6 : 3
	xmin = dec
	xmax = dx - dec
	@xlen = xmax - xmin

	#Current boundaries and value
	vmin = @notify_proc.call(:value_min)
	vmax = @notify_proc.call(:value_max)
	range = vmax - vmin
	current_value = @notify_proc.call(:current_value)
	
	#Management of reversed play
	reversed = @notify_proc.call(:reversed)
	if reversed
		color_gray = 'red'
		color_red = 'gray'
		color_darkred = 'silver'
		color_silver = 'darkred'
	else
		color_gray = 'gray'
		color_red = 'red'
		color_darkred = 'darkred'
		color_silver = 'silver'
	end
		
	#Main line
	line = [Geom::Point3d.new(xmin, dy2), Geom::Point3d.new(xmax, dy2)]
	lst_gl.push [GL_LINE_STRIP, line, color_gray, ywid, '']
	
	#Return lst_gl if not yet initialized	
	return lst_gl unless @xprog && range != 0
	
	#Drawing the elapsed part
	@xprog = xmax if @xprog > xmax
	@xprog = xmin if @xprog < xmin
	ratio = (current_value - vmin) / range	
	xpos = xmin + @xlen * ratio
	if ratio > 0
		line_elapsed = [Geom::Point3d.new(xmin, dy2), Geom::Point3d.new(xpos, dy2)]
		lst_gl.push [GL_LINE_STRIP, line_elapsed, color_red, ywid, '']
	end	

	unless in_zone
		lst_gl.push [GL_LINE_STRIP, line_elapsed, color_red, ywid, '']
		return lst_gl
	end
	
	if @xprog > xpos
		lst_gl.push [GL_LINE_STRIP, line_elapsed, color_red, ywid, '']
		line = [Geom::Point3d.new(xpos, dy2), Geom::Point3d.new(@xprog, dy2)]
		lst_gl.push [GL_LINE_STRIP, line, color_silver, ywid, '']
	elsif @xprog < xpos
		lst_gl.push [GL_LINE_STRIP, line_elapsed, color_darkred, ywid, '']
		line = [Geom::Point3d.new(xmin, dy2), Geom::Point3d.new(@xprog, dy2)]
		lst_gl.push [GL_LINE_STRIP, line, color_red, ywid, '']
	end
	
	radb = 7
	rads = 4
	color = (@xprog && @xprog >= xpos - radb && @xprog <= xpos + radb) ? 'green' : 'red'
	cbig = G6.pts_circle xpos, dy2, radb
	csmall = G6.pts_circle xpos, dy2, rads
	lst_gl.push [GL_POLYGON, cbig, 'white']
	lst_gl.push [GL_POLYGON, csmall, color]
	
	lst_gl
end

#SLIDER: Management of mouse events for the progression bar
def zone_event(code, button_down, dx, dy, x, y)
	@xprog = x
	@yprog = y
	@code_progression = code
	
	#Calculating the ratio
	dec = @progression_dec
	ratio = (x - dec) / (dx - 2 * dec)
	ratio = 1 if ratio > 1
	ratio = 0 if ratio < 0
	@progression_ratio = (code == :out) ? nil : ratio
	
	if @palette.button_down?
		@button_down_progression = true
		@notify_proc.call(:set_value, value_from_ratio(ratio))
	elsif code == :double_click	
		@notify_proc.call(:double_click)
	end
end

#SLIDER: Compute the value from a ratio
def value_from_ratio(ratio=nil)
	val, vround = value_rounding(ratio)
	(vround) ? vround : val
end

#SLIDER: Compute the value from a ratio
def value_rounding(ratio=nil)
	ratio = @progression_ratio unless ratio
	vmin = @notify_proc.call(:value_min)
	vmax = @notify_proc.call(:value_max)
	vincr = @notify_proc.call(:value_ticks)
	range = vmax - vmin
	val = vmin + range * ratio	
	vround = nil
	return [val, vround] unless vincr && @xlen
	
	#Rounding
	if vincr && vincr.class == Array
		vround = dmin = nil
		vincr.each do |v|
			d = (v - val).abs
			if @xlen * 1.0 * d / range.abs < @pix_precision
				if !dmin || d < dmin
					dmin = d
					vround = v
				end	
			end	
		end
		
	elsif vincr.is_a?(Numeric)
		m, d = val.abs.divmod(vincr.abs)
		if @xlen * 1.0 * d / range.abs < @pix_precision
			vround = (val > 0) ? m * vincr : -m * vincr
		end	
	end
	
	#Returning the value
	[val, vround]
end

#SLIDER: Drawing the target time
def draw_time(view, but_symb, xmove, ymove)
	return unless @progression_ratio || @palette.button_hidden?(@button_symb)
	
	#Text for value
	val, vround = value_rounding(@progression_ratio)
	if vround
		val = vround
		frcolor = @fr_color_text_exact
		frwid = 2
	elsif !@notify_proc.call(:value_ticks)
		frcolor = @fr_color_text_exact
		frwid = 2
	else
		frcolor = @fr_color_text
		frwid = 1
	end	
	text = @notify_proc.call(:format_value, val)
	return unless text
	
	w, = G6.simple_text_size(text)
	xt = xmove - w / 2
	vpwidth = view.vpwidth
	if xmove < w
		xt += w/2
	elsif xmove + w > vpwidth	
		xt = vpwidth - w - 3
	end	
	
	yt = ymove + 30
	vpheight = view.vpheight
	yt = ymove - 30 if yt > vpheight - 20
	G6.draw_rectangle_text_at(view, xt, yt, text, @bk_color_text, frcolor, 0, frwid)	
end

#SLIDER: Control method for the slider when based on an Input
def input_slider_proc(code, *args)
	case code
	when :value_min
		@input.get_vmin
	when :value_max	
		@input.get_vmax
	when :value_ticks
		@input.get_vincr
	when :current_value
		@input.get_value
	when :double_click
		@input.dialog_ask
	when :set_value
		value, = args
		@input.set_value value
	when :format_value
		value, = *args
		@input.compute_show_text value
	else
		nil
	end
end
	
end	#class PaletteSliderButton

#*************************************************************************
#*************************************************************************
# TABLE: Management of Table
#*************************************************************************
#*************************************************************************

class PaletteTable

def initialize(palette, *hspecs)
	puts "\nTABLE init"
	@palette = palette
	
	#Default parameters
	@nb_row_min = 2
	@nb_row_max = 4
	@initialized = false
	
	@lst_buttons_row = []
	
	#Parsing the parameters
	hspecs.each do |hspec|
		hspec.each { |key, value| parse_specs(key, value) } if hspec.class == Hash
	end
end

#INIT: Parsing the table specifications
def parse_specs(key, value)
	case key
	when :notify_proc
		@notify_proc = value
	when :nb_row_min
		@nb_row_min = value
	when :nb_row_max
		@nb_row_max = value
	end	
end

#INIT: Invoke the caller notify method
def notify(event, *args)
	@notify_proc.call event, *args
end
	
#INIT: Associate the table to the hosting button
def associate(button)
	@master_button = button
	puts "associate button #{button.symb}"
	initialize_rows
end

def initialize_rows	
	#Initialize the rows
	irow = 0
	while true
		break if irow > 20
		ls_buttons = notify :init_row, irow
		puts "ls buttons = #{ls_buttons.inspect}"
		break unless ls_buttons
		ls_buttons.each_with_index do |lcol, icol|
			puts "column #{icol}"
			lcol.each do |button_specs|
				but_symb = button_specs[0]
				hargs = button_specs[1..-1] + [{ :table_button => @master_button, :table_cell => [irow, icol] }]
				button = @palette.declare_button(but_symb, *hargs)
			end
		end	
		irow += 1
	end
	@initialized = true
end
	
def draw(view, tr)
	pts = G6.pts_rectangle 0, 0, 50, 80
	view.drawing_color = 'green'
	view.draw2d GL_POLYGON, pts.collect { |pt| tr * pt }
end
	
end	#class PaletteTable

end	#module Traductor

