=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed Apr. 2015 by Fredo6

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:  body_Lib6ViewPoint.rb
# Original Date	:  17 Apr 2015
# Type			:  Script library part of the LibFredo6 shared libraries (implemnetation)
# Description	:  A utility library to manage viewpoint (views or pages, with style and other options
#-------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module Traductor

#--------------------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------------------
# class ViewPointManager: Viewpoint Management
#--------------------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------------------

class ViewPointManager

#Class instance Initialization
def initialize__(*hargs)
	@model = Sketchup.active_model
	@view = @model.active_view
	
	@struct_ViewPoint = Struct.new :type, :sucamera, :rendering
end

#Calculate the viewpoint for the current view
def viewpoint_create(type, sucamera=nil, rendering=nil)
	vp = @struct_ViewPoint.new
	vp.type = type
	vp.sucamera = sucamera
	vp.rendering = {}
	vp.rendering.update rendering if rendering
	vp
end

#Calculate the viewpoint for the current view
def viewpoint_current(rendering=nil)
	viewpoint_create(:view, G6.camera_duplicate(@view.camera), rendering)
end

#Calculate the viewpoint for a specified page
def viewpoint_page(page, rendering=nil)
	vp = @struct_ViewPoint.new
	vp.type = :page
	vp.sucamera = page.camera
	vp.rendering = {}
	vp.rendering.update rendering if rendering
	vp
end

#Switch the model view to a viepoint
def switch_to_viewpoint(vp)
	sucamera = G6.camera_duplicate(vp.sucamera)
	@view.camera = sucamera
end

#Take a picture of the current view
def take_picture_current_view(file, px=nil, py=nil)
	ratio = 1.0 * @view.vpheight / @view.vpwidth
	if px
		py = (px * ratio).round unless py
	elsif py
		px = (py / ratio).round unless px
	else
		px = @view.vpwidth
		py = @view.vpheight
	end
	@view.write_image file, px, py, true
	[px, py]
end

#Take a picture of a specified page
def take_picture_viewpoint(vp, file, px, py=nil)
	vp_cur = viewpoint_current
	switch_to_viewpoint vp
	px, py = take_picture_current_view file, px, py
	switch_to_viewpoint vp_cur
	[px, py]
end

#Take a picture of a specified page
def take_picture_page(page, file, px, py=nil)
	vp_cur = viewpoint_current
	vp_page = viewpoint_page(page)
	switch_to_viewpoint vp_page
	px, py = take_picture_current_view file, px, py
	switch_to_viewpoint vp_cur
	[px, py]
end

end	#class ViewPointManager

end #Module Traductor
