=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed Apr. 2015 by Fredo6

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:  Lib6TabDialog.rb
# Original Date	:  16 Apr 2015
# Type			:  Script library part of the LibFredo6 shared libraries
# Description	:  A utility library to assist Javascript Tab dialog
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module Traductor

#--------------------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------------------
# class HTML_TabDialog: Expandable table
#--------------------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------------------

class HTML_TabDialog

#Initialize the class instance
def initialize__(tabdlg_id, html, wdlg)
	@tabdlg_id = tabdlg_id
	@content_id = "TabDlg_#{@tabdlg_id}"
	@html = html
	@wdlg = wdlg
	@lst_tabs = []
	@hsh_tabs = {}
	@tab_active_id = nil
	@wdlg.register_tab_dialog self, tabdlg_id
	
	@default_bkcolor = 'lightgrey'
	
	@border_color = 'darkgray'
	@radius = '10px'
	@space_tab = 5	
end

#Register the tabs
def register_tab(hparams)
	id = hparams[:id]
	@lst_tabs.push hparams
	@hsh_tabs[id] = hparams
	hparams
end

def declare_active(id)
	@tab_active_id = id	
end

def declare_reference_tab(id)
	@tab_ref_id = id	
end


#Computation of ID for HTML elements
def get_div_id(tab_id) ; "TabDlg_Div_#{@tabdlg_id}_#{tab_id}" ; end
def get_td_id(tab_id) ; "TabDlg_Td_#{@tabdlg_id}_#{tab_id}" ; end
def get_dlg_cell_id ; "TabDlg_Cell_#{@tabdlg_id}" ; end

#Handle the events (update main table) and notify of clients
def notify_event(event, type, svalue)
	case type
	when /SwitchRequest/i
		new_tab_id = svalue
		return if new_tab_id == @tab_active_id
		deny = @wdlg.notify_callback("TabDialog", "RequestSwitch", @tabdlg_id, new_tab_id) 
		return if deny == :deny
		switch_tab new_tab_id
		@wdlg.notify_callback("TabDialog", "AfterSwitch", @tabdlg_id, new_tab_id) 
	end	
end

#Switch the active_tab
def switch_tab(new_tab_id)
	htab_new = @hsh_tabs[new_tab_id]
	return unless htab_new
	cur_tab_id = @tab_active_id
	color_new = htab_new[:bkcolor]
	
	#Initializing the switch
	script = ""
	
	#Switching the color
	cell_id = get_dlg_cell_id
	script += "document.getElementById(\"#{cell_id}\").style.background = '#{color_new}' ;"
	
	#Switching the bottom border color
	td_id_cur = get_td_id(cur_tab_id)
	td_id_new = get_td_id(new_tab_id)
	script += "document.getElementById(\"#{td_id_new}\").style.borderBottomColor = '#{color_new}' ;"
	script += "document.getElementById(\"#{td_id_cur}\").style.borderBottomColor = '#{@border_color}' ;"
	
	#Switching the visibility of dialog contents
	div_id_cur = get_div_id(cur_tab_id)
	div_id_new = get_div_id(new_tab_id)
	script += "document.getElementById(\"#{div_id_cur}\").style.visibility = 'hidden' ;"
	script += "document.getElementById(\"#{div_id_new}\").style.visibility = 'visible' ;"
	
	#Setting the new tab active and performing the switch
	@tab_active_id = new_tab_id
	@wdlg.execute_script script
end

#Generate the HTML
def generate_html
	@tab_active_id = @lst_tabs[0][:id] unless @tab_active_id
	
	ntab = @lst_tabs.length
	ncell = 2 * ntab
	
	#Style sheet contribution for rounding corners of cells
	rounding = "-webkit-border-top-left-radius: #{@radius}; -webkit-border-top-right-radius: #{@radius};"
	rounding += "-moz-border-radius-topleft: #{@radius}; -moz-border-radius-topright: #{@radius};"
	rounding += "border-top-left-radius: #{@radius}; border-top-right-radius: #{@radius}; }"
	
	#Generate the Tab table
	text = ""
	cdec = 2 * @space_tab + 2
	style_tab_table = "position: relative ; z-index: 3 ;"
	text += "<table width='auto' cellpadding='0px' cellspacing='#{@space_tab}px' style='#{style_tab_table}' >"
	
	#Creating the tabs as cells of the Tab table
	text += "<tr>"
	@lst_tabs.each do |htab|
		html_tab = htab[:html_tab]
		bkcolor = htab[:bkcolor]
		w = 2
		tab_id = htab[:id]
		if tab_id == @tab_active_id
			bcolor = bkcolor
		else
			bcolor = @border_color
		end	
		action = "onClick='TabDlg_switch_request(\"#{@tabdlg_id}\", \"#{tab_id}\");'"
		style = "border: #{w}px solid #{@border_color} ; border-bottom: #{w}px solid #{bcolor} ; "
		style += "padding: 2px 8px ; background: #{bkcolor} ; cursor: pointer ;"
		style += rounding
		tip = (htab[:tip]) ? htab[:tip] : ""
		td_id = get_td_id(tab_id)
		wid = (htab[:width]) ? "width='#{htab[width]}'" : ""
		text += "<td ID='#{td_id}' #{wid} title='#{tip}' align='left' style='#{style}' #{action}>#{html_tab}</td>"
	end
	text += "</tr></table>"
	
	#Creating the dialog area for each tab as row of the dialog table (only one visible)
	hdiv_h = ""
	hdiv = ""
	dec = 3 * @space_tab
	@lst_tabs.each do |htab|
		style_div = "position: absolute ; top: #{dec}px ; left: #{dec}px ;"
		style_div += "visibility: hidden ;" unless htab[:id] == @tab_active_id
		div_id = get_div_id(htab[:id])
		hdiv += "<div ID='#{div_id}' style='#{style_div}'>#{htab[:html_dialog]}</div>"	
	end
	
	#Creating the place holder for reference tab - Remove ID for avoiding conflicts in HTML
	htab = @hsh_tabs[@tab_ref_id]
	htab = @lst_tabs[0] unless htab
	txt_h = htab[:html_dialog].gsub(/\sID='/i, " ID='AA_")
	hdiv_h += "<div style='visibility:hidden ; float: left'>#{txt_h}</div>"
	
	#Creating the Dialog areas as a second table 
	style_dialog_table = "position: relative ; top: -#{cdec}px ; z-index: 2 ; padding: #{@space_tab}px ;"
	htab = @hsh_tabs[@tab_active_id]
	bkcolor = htab.fetch :bkcolor, @default_bkcolor
	style = "border: 2px solid #{@border_color} ; padding: 8px 8px ; background: #{bkcolor} ;"
	td_id = get_dlg_cell_id
	text += "<table style='#{style_dialog_table}' width='100%' cellspacing='0x' >"
	text += "<td ID='#{td_id}' style='#{style}'>#{hdiv_h}#{hdiv}</td>"
	text += "</table>"
	
	text
end

#Specific scripts for Tabbed dialog
def special_scripts(html)

#Main functions for Expand and Shrink
	text = %Q~

function TabDlg_switch_request(dlg_id, id_to) { 
	SUCallback ("Action", "TabDialog", "SwitchRequest", dlg_id, massage(id_to)) ; 
}

~
	
	html.script_add text
end

end	#class HTML_TabDialog

end #Module Traductor
